import os
import sys
import time
import networkx as nx
import community
import shutil

__author1__ = "Sucheta Soundarajan"
__author2__ = "John Hopcroft"
__contact__ = "susounda@syr.edu"


def timeit(method):
    """
    Decorator: Compute the execution time of a function
    :param method: the function
    :return: the method runtime in seconds
    """

    def timed(*arguments, **kw):
        ts = time.time()
        result = method(*arguments, **kw)
        te = time.time()
        sys.stdout.write("------------------------------------\n")
        sys.stdout.write(
            "Time:  %r %2.2f sec\n" % (method.__name__.strip("_"), te - ts)
        )
        sys.stdout.write("------------------------------------\n")
        sys.stdout.flush()
        return result

    return timed


class NodePerception(object):
    """
    Node Perception

    Sucheta Soundarajan and John Hopcroft.
    Use of Local Group Information to Identify Communities in Networks.
    ACM Transactions on Knowledge Discovery from Data (TKDD). 2015.

    http://www.soundarajan.org/papers/TKDDCommunities.pdf

    """

    def __init__(
        self, g, sim_threshold=0.2, overlap_threshold=1, min_comm_size=2, out_file=""
    ):
        """
        Constructor

        :param network_filename: the .ncol network file
        :param sim_threshold: the tolerance required in order to merge communities
        :param overlap_threshold: the overlap tolerance
        :param min_comm_size: minimum algorithms size
        :param out_file: desired output file name
        """
        self.g = g
        self.out_file = out_file
        self.sim_threshold = sim_threshold
        self.min_comm_size = min_comm_size
        self.overlap_threshold = overlap_threshold

        if not os.path.exists("tmp"):
            os.makedirs("tmp")

    def __GetEdgesIntoDict(self):
        edges = {}
        for t in self.g.edges():
            if len(t) > 0:
                if t[0] != t[1]:
                    if t[0] not in edges:
                        edges[t[0]] = {t[1]}
                    else:
                        edges[t[0]].add(t[1])
                    if t[1] not in edges:
                        edges[t[1]] = {t[0]}
                    else:
                        edges[t[1]].add(t[0])
        return edges

    def __FirstPartition(self, edges, first_part_file):
        OUT = open(first_part_file, "w")
        node_count = 0
        for n in edges:
            node_count = node_count + 1
            if node_count > 0:
                # the nodes
                index = {}
                reverse_index = {}
                count = 0
                to_add_edges = []
                adj = set([])
                for neighbor in edges[n]:
                    index[count] = neighbor
                    reverse_index[neighbor] = count
                    adj.add(neighbor)
                    count = count + 1
                for m in reverse_index:
                    for k in edges[m]:
                        if k in reverse_index and reverse_index[k] < reverse_index[m]:
                            to_add_edges.append((reverse_index[m], reverse_index[k]))
                G = nx.Graph()
                G.add_nodes_from([i for i in index])
                G.add_edges_from(to_add_edges)
                if len(to_add_edges) > 0:
                    dict_H = community.best_partition(G)
                    H = {}
                    for node in dict_H:
                        if dict_H[node] not in H:
                            H[dict_H[node]] = set([])
                        H[dict_H[node]].add(node)

                    for i in H:
                        comm = H[i]
                        if len(comm) > 0:
                            for c in comm:
                                OUT.write(str(index[int(c)]) + " ")
                            OUT.write(str(n))
                            OUT.write("\n")
                        elif len(comm) > 0:
                            for c in comm:
                                if index[int(c)] in edges[n]:
                                    OUT.write(str(index[int(c)]) + " ")
                            OUT.write(str(n))
                            OUT.write("\n")
        OUT.close()

    def __GetMembership(self, first_part_file, membership_file):
        node_membership = {}

        IN = open(first_part_file, "rb")
        read_line = IN.readline()
        count = 0
        while read_line:
            t = read_line.rstrip().split()
            if len(t) >= self.min_comm_size:
                for mem in t:
                    if mem not in node_membership:
                        node_membership[mem] = set([])
                    node_membership[mem].add(count)
            count = count + 1
            read_line = IN.readline()

        IN.close()

        OUT = open(membership_file, "w")
        for n in node_membership:
            in_comms = node_membership[n]
            OUT.write(str(n) + " ")
            for c in in_comms:
                OUT.write(str(c) + " ")
            OUT.write("\n")
        OUT.close()

    def __Jaccard(self, set1, set2):
        set1 = set(set1)
        set2 = set(set2)
        return float(len(set1.intersection(set2))) / float(len(set1.union(set2)))

    def __GetCommSimilarities(self, first_part_file, membership_file, sim_file):

        IN = open(first_part_file, "rb")
        line_offset = {}
        offset = 0
        count = 0
        for line in IN:
            line_offset[count] = offset
            count = count + 1
            offset += len(line)
        IN.close()
        num_lines = count - 1

        IN = open(membership_file, "r")
        node_membership = {}
        read_line = IN.readline()
        while read_line:
            t = read_line.rstrip().split()

            if len(t) > 0:
                node = t[0]
                t.remove(node)
                if node not in node_membership:
                    node_membership[node] = set([])
                for c in t:
                    node_membership[node].add(c)
            read_line = IN.readline()

        IN.close()
        IN = open(first_part_file, "rb")
        OUT = open(sim_file, "w")
        curr_count = 0
        total_lines = 0

        while curr_count < num_lines:
            IN.seek(line_offset[curr_count])
            read_line = IN.readline()
            t = read_line.rstrip().split()
            if len(t) >= self.min_comm_size:
                adj = set([])
                for mem in t:
                    adj = adj.union(set(node_membership[str(mem)]))
                for adj_comm in adj:
                    if int(adj_comm) > int(curr_count):
                        IN.seek(line_offset[int(adj_comm)])
                        read_line = IN.readline()
                        r = read_line.rstrip().split()
                        if len(r) >= self.min_comm_size:
                            sim = self.__Jaccard(t, r)
                            if r[len(r) - 1] == t[len(t) - 1]:
                                sim = 0
                            if sim > self.sim_threshold:
                                total_lines = total_lines + 1
                                OUT.write(
                                    str(curr_count)
                                    + " "
                                    + str(adj_comm)
                                    + " "
                                    + str(sim)
                                    + " "
                                    + str(len(set(t).intersection(set(r))))
                                    + " "
                                    + str(len(t))
                                    + " "
                                    + str(len(r))
                                )
                                OUT.write("\n")
            curr_count = curr_count + 1
        IN.close()
        OUT.close()

    def __SecondPartition(self, overlap_threshold, sim_file, first_part_file):
        return_vals = self.__ModClusteringSingleBig(
            sim_file, first_part_file, overlap_threshold
        )
        return return_vals

    def __ModClusteringSingleBig(self, sim_file, first_part_file, overlap_threshold=0):
        IN = open(sim_file, "r")
        read_line = IN.readline()
        num_lines = 0
        comm_edges = {}
        to_add_edges = []
        while read_line:
            t = read_line.rstrip().split()
            num_lines += 1
            if len(t) > 0:
                node1 = int(t[0])
                node2 = int(t[1])
                sim = t[2]
                overlap = t[3]
                if (
                    float(sim) >= self.sim_threshold
                    and float(overlap) > overlap_threshold
                ):
                    if node1 not in comm_edges:
                        comm_edges[node1] = set([])
                    if node2 not in comm_edges:
                        comm_edges[node2] = set([])
                    if node2 not in comm_edges[node1]:
                        comm_edges[node1].add(node2)
                        comm_edges[node2].add(node1)
                        weight = sim
                        to_add_edges.append((node1, node2, {"weight": float(weight)}))

            read_line = IN.readline()
        IN.close()
        G = nx.Graph()
        G.add_nodes_from(range(len(comm_edges)))
        G.add_edges_from(to_add_edges)
        dict_H = community.best_partition(G)
        H1 = {}
        for e in dict_H:
            if dict_H[e] not in H1:
                H1[dict_H[e]] = set([])
            H1[dict_H[e]].add(e)
        H = []
        for e in H1:
            H.append(H1[e])

        IN = open(first_part_file, "rb")
        line_offset = {}
        offset = 0
        count = 0
        for line in IN:
            line_offset[count] = offset
            count = count + 1
            offset += len(line)
        IN.close()

        IN = open(first_part_file, "rb")
        all_comms = {}
        i = 0
        for big_comm in H:
            comm_members = {}
            for comm in big_comm:
                IN.seek(line_offset[int(comm)])
                read_line = IN.readline()
                t = read_line.rstrip().split()
                if len(t) > 0:
                    for t1 in t:
                        if t1 not in comm_members:
                            comm_members[t1] = 0
                        comm_members[t1] += 1
            all_comms[i] = set([])
            for t1 in comm_members:
                if comm_members[t1] >= 0:
                    all_comms[i].add(t1)

            i += 1

        return all_comms

    def __GetModComms(self, G):
        dict_H = community.best_partition(G)
        H1 = {}
        for e in dict_H:
            if dict_H[e] not in H1:
                H1[dict_H[e]] = set([])
            H1[dict_H[e]].add(e)
        H = []
        for e in H1:
            H.append(H1[e])
        return H

    def __CleanComms(self, to_clean):
        comms = {}

        count = 0
        idx = {}
        for t in to_clean.values():
            if len(t) > 0:
                comms[count] = set(t)
                for i in t:
                    if i not in idx:
                        idx[i] = set([])

                    idx[i].add(count)
                count += 1
            elif len(t) > 0:
                comms[count] = set(t)
                count += 1

        coms = []
        for i in range(count):
            C = comms[i]
            if len(C) > 0:
                poss = set([])
                found = 0
                for n in C:
                    poss = poss.union(idx[n])
                for j in poss:
                    if j < i:
                        if (
                            len(comms[j]) == len(comms[i])
                            and len(comms[j].difference(comms[i])) == 0
                        ):
                            found = 1
                if found != 1:
                    coms.append([t.decode("utf-8") for t in C])
            else:
                coms.append([t.decode("utf-8") for t in C])
        return coms

    def execute(self):

        first_part_file = "tmp/part1_"
        membership_file = "tmp/membership_"
        sim_file = "tmp/simfile_"

        # Part 1
        # read the edge file
        edges = self.__GetEdgesIntoDict()

        sys.stdout.write("First Partition\n")
        # get the small communities and write them to a file
        self.__FirstPartition(edges, first_part_file)

        self.__GetMembership(first_part_file, membership_file)
        # calculate similarity between each pair of small communities; minimum threshold is set to jaccard of 0.2
        self.__GetCommSimilarities(first_part_file, membership_file, sim_file)

        # Part 2
        sys.stdout.write("Second Partition\n")
        return_vals = self.__SecondPartition(0, sim_file, first_part_file)
        coms = self.__CleanComms(return_vals)

        shutil.rmtree("tmp")
        return coms



import networkx as nx
import random
from networkx.algorithms import community as com
import random
import collections


def __bfs(graph, root, epsilon):
    seen, queue, distance = {root}, collections.deque([root]), 1
    while queue:
        parent = queue.popleft()
        for child in nx.all_neighbors(graph, parent):
            if child not in seen:
                if distance < epsilon:
                    seen.add(child)
                    queue.append(child)
        distance += 1
    return seen


def endntm_evalFuction(graph, clusters_list, etha=0.5):
    mod = com.modularity(graph, clusters_list)
    coverage = nx.algorithms.community.quality.coverage(graph, clusters_list)
    val = (1 - etha) * coverage + etha * mod
    return val


def endntm_find_overlap_cluster(graph, clusters_list, epsilon):

    node_cluster_dic = {}
    cluster_node_dic = {}
    for i, cluster in zip(range(len(clusters_list)), clusters_list):
        for node in cluster:
            node_cluster_dic[node] = i + 1
        cluster_node_dic[i + 1] = set(cluster)

    cluster_no = 0

    # finding Overlapping candidate node
    overlap_cand_dic = {}
    for cluster in clusters_list:
        cluster_no += 1
        for node in cluster:
            neighborhood = __bfs(graph, node, epsilon)
            neighborhood = neighborhood - cluster
            if list(neighborhood) != list():
                overlap_cand_dic[node] = neighborhood

    # filtering overlapping node
    for node in list(overlap_cand_dic.keys()):
        nghbr_clstr_dic = {}
        neighborhood = overlap_cand_dic[node]
        for vertex in list(neighborhood):
            if node_cluster_dic[vertex] not in list(nghbr_clstr_dic):
                nghbr_clstr_dic[node_cluster_dic[vertex]] = {vertex}
            else:
                nghbr_clstr_dic[node_cluster_dic[vertex]].add(vertex)

        # calculating distribution threshold
        disb_threshold = graph.degree(node) / (len(nghbr_clstr_dic) + 1)

        # Generating the overlapping cluster by including overlapping node
        for cid in list(nghbr_clstr_dic):
            if len(nghbr_clstr_dic[cid]) >= disb_threshold:
                cluster_node_dic[cid].add(node)

    overlap_cluster = list(
        cluster_node_dic[cid] for cid in list(cluster_node_dic.keys())
    )
    return overlap_cluster



import numpy as np
import copy
from collections import defaultdict


def slpa_nx(G, T, r):
    """
    Speaker-Listener Label Propagation Algorithm (SLPA)
    see http://arxiv.org/abs/1109.5720
    """

    # Stage 1: Initialization
    memory = {i: {i: 1} for i in G.nodes()}

    # Stage 2: Evolution
    for t in range(T):

        listeners_order = list(G.nodes())
        np.random.shuffle(listeners_order)

        for listener in listeners_order:
            speakers = G[listener].keys()
            if len(speakers) == 0:
                continue

            labels = defaultdict(int)

            for j, speaker in enumerate(speakers):
                # Speaker Rule
                total = float(sum(memory[speaker].values()))
                labels[
                    list(memory[speaker].keys())[
                        np.random.multinomial(
                            1, [freq / total for freq in memory[speaker].values()]
                        ).argmax()
                    ]
                ] += 1

            # Listener Rule
            accepted_label = max(labels, key=labels.get)

            # Update listener memory
            if accepted_label in memory[listener]:
                memory[listener][accepted_label] += 1
            else:
                memory[listener][accepted_label] = 1

    # Stage 3:
    for node, mem in memory.items():
        its = copy.copy(list(mem.items()))
        for label, freq in its:
            if freq / float(T + 1) < r:
                del mem[label]

    # Find nodes membership
    communities = {}
    for node, mem in memory.items():
        for label in mem.keys():
            if label in communities:
                communities[label].add(node)
            else:
                communities[label] = {node}

    # Remove nested communities
    nested_communities = set()
    keys = list(communities.keys())
    for i, label0 in enumerate(keys[:-1]):
        comm0 = communities[label0]
        for label1 in keys[i + 1 :]:
            comm1 = communities[label1]
            if comm0.issubset(comm1):
                nested_communities.add(label0)
            elif comm0.issuperset(comm1):
                nested_communities.add(label1)

    for comm in nested_communities:
        del communities[comm]

    return [list(c) for c in communities.values()]



import copy
import networkx as nx

__author__ = "Giulio Rossetti"
__contact__ = "giulio.rossetti@gmail.com"
__website__ = "about.giuliorossetti.net"
__license__ = "BSD"


class eTILES(object):
    """
    TILES
    Algorithm for evolutionary community discovery
    ***Explicit removal***
    """

    def __init__(self, dg: object, obs: int = 7):
        """
        Constructor
        :param g: DyNetx graph
        :param obs: observation window
        :param start: starting date
        :param end: ending date
        """
        self.cid = 0
        self.actual_slice = obs
        self.dg = dg
        self.g = nx.Graph()
        self.splits = None
        self.removed = 0
        self.added = 0
        self.obs = obs
        self.communities = {}
        self.mathces = []

    @property
    def new_community_id(self) -> int:
        """
        Return a new community identifier
        :return: new community id
        """
        self.cid += 1
        self.communities[self.cid] = {}
        return self.cid

    def execute(self) -> dict:
        """
        Execute TILES algorithm
        """

        last_break = min(self.dg.temporal_snapshots_ids())
        count = 0

        #################################################
        #                   Main Cycle                  #
        #################################################

        for l in self.dg.stream_interactions():
            e = {}
            action = l[2]
            u = l[0]
            v = l[1]
            dt = int(l[3])

            e["weight"] = 1
            e["u"] = u
            e["v"] = v

            #############################################
            #               Observations                #
            #############################################

            dif = dt - last_break

            if dif >= self.obs:
                last_break = dt
                self.added -= 1
                self.added = 1
                self.removed = 0
                self.print_communities()
                yield self.communities

            if u == v:
                continue

            if action == "-":
                self.remove_edge(e)
                continue

            if not self.g.has_node(u):
                self.g.add_node(u)
                self.g.nodes[u]["c_coms"] = {}

            if not self.g.has_node(v):
                self.g.add_node(v)
                self.g.nodes[v]["c_coms"] = {}

            if self.g.has_edge(u, v):
                w = self.g.adj[u][v]["weight"]
                self.g.adj[u][v]["weight"] = w + e["weight"]
                continue
            else:
                self.g.add_edge(u, v, weight=e["weight"])

            u_n = list(self.g.neighbors(u))
            v_n = list(self.g.neighbors(v))

            #############################################
            #               Evolution                   #
            #############################################

            # new community of peripheral nodes (new nodes)
            if len(u_n) > 1 and len(v_n) > 1:
                common_neighbors = set(u_n) & set(v_n)
                self.common_neighbors_analysis(u, v, common_neighbors)

            count += 1

        #  Last writing
        self.added = 0
        self.removed = 0
        self.print_communities()

        yield self.communities

    def print_communities(self):
        """
        Print the actual communities
        """

        nodes_to_coms = {}
        merge = {}
        coms_to_remove = []
        drop_c = []

        for idc, comk in self.communities.items():

            com = comk.keys()

            if self.communities[idc] is not None:
                if len(com) > 2:
                    key = tuple(sorted(com))

                    # Collision check and merge index build (maintaining the lowest id)
                    if key not in nodes_to_coms:
                        nodes_to_coms[key] = idc
                    else:
                        old_id = nodes_to_coms[key]
                        drop = idc
                        if idc < old_id:
                            drop = old_id
                            nodes_to_coms[key] = idc

                        # merged to remove
                        coms_to_remove.append(drop)
                        if not nodes_to_coms[key] in merge:
                            merge[nodes_to_coms[key]] = [idc]
                        else:
                            merge[nodes_to_coms[key]].append(idc)
                else:
                    drop_c.append(idc)
            else:
                drop_c.append(idc)

        for dc in drop_c:
            self.destroy_community(dc)

        # Community Cleaning
        for comid, c_val in merge.items():
            # maintain minimum community after merge
            c_val.append(comid)
            k = min(c_val)

            c_val.remove(k)
            if self.actual_slice > self.obs:
                for fr in c_val:
                    self.mathces.append(
                        (
                            f"{self.actual_slice-self.obs}_{fr}",
                            f"{self.actual_slice}_{k}",
                            None,
                        )
                    )

        m = 0
        for c in coms_to_remove:
            self.destroy_community(c)
            m += 1

        self.actual_slice += self.obs

    def common_neighbors_analysis(self, u, v, common_neighbors):
        """
        General case in which both the nodes are already present in the net.
        :param u: a node
        :param v: a node
        :param common_neighbors: common neighbors of the two nodes
        """

        # no shared neighbors
        if len(common_neighbors) < 1:
            return

        else:

            shared_coms = set(self.g.nodes[v]["c_coms"].keys()) & set(
                self.g.nodes[u]["c_coms"].keys()
            )
            only_u = set(self.g.nodes[u]["c_coms"].keys()) - set(
                self.g.nodes[v]["c_coms"].keys()
            )
            only_v = set(self.g.nodes[v]["c_coms"].keys()) - set(
                self.g.nodes[u]["c_coms"].keys()
            )

            # community propagation: a community propagates iff at least two of [u, v, z] are central
            propagated = False

            for z in common_neighbors:
                for c in self.g.nodes[z]["c_coms"].keys():
                    if c in only_v:
                        self.add_to_community(u, c)
                        propagated = True

                    if c in only_u:
                        self.add_to_community(v, c)
                        propagated = True

                for c in shared_coms:
                    if c not in self.g.nodes[z]["c_coms"]:
                        self.add_to_community(z, c)
                        propagated = True

            else:
                if not propagated:
                    # new community
                    actual_cid = self.new_community_id
                    self.add_to_community(u, actual_cid)
                    self.add_to_community(v, actual_cid)

                    for z in common_neighbors:
                        self.add_to_community(z, actual_cid)

    def remove_edge(self, e):
        """
        Edge removal procedure
        :param e: edge
        """

        coms_to_change = {}

        self.removed += 1
        u = e["u"]
        v = e["v"]

        if self.g.has_edge(u, v):

            # u and v shared communities
            if (
                len(list(self.g.neighbors(u))) > 1
                and len(list(self.g.neighbors(v))) > 1
            ):
                coms = set(self.g.nodes[u]["c_coms"].keys()) & set(
                    self.g.nodes[v]["c_coms"].keys()
                )

                for c in coms:
                    if c not in coms_to_change:
                        cn = set(self.g.neighbors(u)) & set(self.g.neighbors(v))
                        coms_to_change[c] = [u, v]
                        coms_to_change[c].extend(list(cn))
                    else:
                        cn = set(self.g.neighbors(u)) & set(self.g.neighbors(v))
                        coms_to_change[c].extend(list(cn))
                        coms_to_change[c].extend([u, v])
                        ctc = set(coms_to_change[c])
                        coms_to_change[c] = list(ctc)
            else:
                if len(list(self.g.neighbors(u))) < 2:
                    coms_u = copy.copy(list(self.g.nodes[u]["c_coms"].keys()))
                    for cid in coms_u:
                        self.remove_from_community(u, cid)

                if len(list(self.g.neighbors(v))) < 2:
                    coms_v = copy.copy(list(self.g.nodes[v]["c_coms"].keys()))
                    for cid in coms_v:
                        self.remove_from_community(v, cid)

            self.g.remove_edge(u, v)

        # update of shared communities
        self.update_shared_coms(coms_to_change)

    def destroy_community(self, cid):
        nodes = [x for x in self.communities[cid].keys()]
        for n in nodes:
            self.remove_from_community(n, cid)
        self.communities.pop(cid, None)

    def add_to_community(self, node, cid):

        self.g.nodes[node]["c_coms"][cid] = None
        if cid in self.communities:
            self.communities[cid][node] = None
        else:
            self.communities[cid] = {node: None}

    def remove_from_community(self, node, cid):
        if cid in self.g.nodes[node]["c_coms"]:
            self.g.nodes[node]["c_coms"].pop(cid, None)
            if cid in self.communities and node in self.communities[cid]:
                self.communities[cid].pop(node, None)

    def update_shared_coms(self, coms_to_change):
        # update of shared communities
        for c in coms_to_change:
            if c not in self.communities:
                continue

            c_nodes = self.communities[c].keys()

            if len(c_nodes) > 3:

                sub_c = self.g.subgraph(c_nodes)
                c_components = nx.number_connected_components(sub_c)

                # unbroken community
                if c_components == 1:
                    to_mod = sub_c.subgraph(coms_to_change[c])
                    self.modify_after_removal(to_mod, c)

                # broken community: bigger one maintains the id, the others obtain a new one
                else:
                    new_ids = []

                    first = True
                    components = nx.connected_components(sub_c)
                    for com in components:
                        if first:
                            if len(com) < 3:
                                self.destroy_community(c)
                            else:
                                to_mod = list(set(com) & set(coms_to_change[c]))
                                sub_c = self.g.subgraph(to_mod)
                                self.modify_after_removal(sub_c, c)
                            first = False

                        else:
                            if len(com) > 3:
                                # update the memberships: remove the old ones and add the new one
                                to_mod = list(set(com) & set(coms_to_change[c]))
                                sub_c = self.g.subgraph(to_mod)

                                central = self.centrality_test(sub_c).keys()
                                if len(central) >= 3:
                                    actual_id = self.new_community_id
                                    new_ids.append(actual_id)
                                    for n in central:
                                        self.add_to_community(n, actual_id)

                    # splits
                    if len(new_ids) > 0 and self.actual_slice > 0:
                        for n in new_ids:
                            self.mathces.append(
                                (
                                    f"{self.actual_slice-self.obs}_{c}",
                                    f"{self.actual_slice}_{n}",
                                    None,
                                )
                            )
            else:
                self.destroy_community(c)

    def modify_after_removal(self, sub_c, c):
        """
        Maintain the clustering coefficient invariant after the edge removal phase
        :param sub_c: sub-community to evaluate
        :param c: community id
        """
        central = self.centrality_test(sub_c).keys()

        # in case of previous splits, update for the actual nodes
        remove_node = set(self.communities[c].keys()) - set(sub_c.nodes())

        for rm in remove_node:
            self.remove_from_community(rm, c)

        if len(central) < 3:
            self.destroy_community(c)
        else:
            not_central = set(sub_c.nodes()) - set(central)
            for n in not_central:
                self.remove_from_community(n, c)

    def centrality_test(self, subgraph):
        central = {}

        for u in subgraph.nodes():
            if u not in central:
                cflag = False
                neighbors_u = set(self.g.neighbors(u))
                if len(neighbors_u) > 1:
                    for v in neighbors_u:
                        if u > v:
                            if cflag:
                                break
                            else:
                                neighbors_v = set(self.g.neighbors(v))
                                cn = neighbors_v & neighbors_v
                                if len(cn) > 0:
                                    central[u] = None
                                    central[v] = None
                                    for n in cn:
                                        central[n] = None
                                    cflag = True
        return central

    def get_matches(self):
        return self.mathces



# Author: True Price <jtprice@cs.unc.edu>

from math import log
from collections import defaultdict
import functools


def graphentropy(g, weight=None):
    if weight is None:
        return __graph_entropy(g)
    else:
        return __graph_entropy_w(g, weight)


# define entropy of a vertex in terms of an associated cluster
def __entropy(c, v, data):
    neighbors = data[v]
    inner = len(c & neighbors) / float(len(neighbors))  # p(inner links)
    return (
        0
        if inner == 0 or inner == 1
        else -inner * log(inner, 2) - (1 - inner) * log(1 - inner, 2)
    )


def __entropy_w(c, v, degree, data):
    deg = degree[v]
    if deg == 0:  # if node degree is zero, do nothing
        return 0

    neighbors = data[v]
    # calculate p(inner links)
    inner = sum(w for n, w in neighbors.items() if n in c) / deg
    return (
        0
        if inner <= 0.0 or inner >= 1.0
        else -inner * log(inner, 2) - (1 - inner) * log(1 - inner, 2)
    )


def __graph_entropy(g):
    data = defaultdict(set)  # node id => neighboring node ids

    # read in graph
    for a, b in g.edges():
        data[a].add(b)
        data[b].add(a)

    candidates = set(data)
    clusters = []
    while candidates:
        v = candidates.pop()  # select a random vertex
        cluster = data[v].copy()  # add neighbors to cluster
        cluster.add(v)
        entropies = dict((x, __entropy(cluster, x, data)) for x in data)

        # step 2: try removing neighbors to minimize entropy
        for n in list(cluster):
            if n == v:
                continue  # don't remove our seed, obviously
            new_c = cluster.copy()
            new_c.remove(n)
            new_e = dict((x, __entropy(new_c, x, data)) for x in data[n])
            # if removing the neighbor decreases new entropy (for the node and
            # all its neighbors), then do so
            if sum(new_e.values()) < sum(entropies[x] for x in data[n]):
                cluster = new_c
                entropies.update(new_e)

        # boundary candidates
        c = functools.reduce(lambda a, b: a | b, (data[x] for x in cluster)) - cluster
        while c:
            n = c.pop()
            new_c = cluster.copy()
            new_c.add(n)
            new_e = dict((x, __entropy(new_c, x, data)) for x in data[n])
            if sum(new_e.values()) < sum(entropies[x] for x in data[n]):
                cluster = new_c
                entropies.update(new_e)
                c &= data[n] - cluster

        # remove the elements of the cluster from our candidate set; add cluster
        candidates -= cluster
        clusters.append(list(cluster))

    return clusters


def __graph_entropy_w(g, wlabel):
    data = defaultdict(dict)  # node id => neighboring node id => edge weight
    degree = defaultdict(float)  # weighted node degrees

    # read in graph
    for a, b, dt in g.edges(data=True):
        w = dt[wlabel]
        data[a][b] = w
        data[b][a] = w
        degree[a] += w
        degree[b] += w

    candidates = set(data)
    clusters = []
    while candidates:
        v = candidates.pop()  # select a random vertex
        cluster = set(data[v])  # add neighbors to cluster
        cluster.add(v)
        entropies = dict((x, __entropy_w(cluster, x, degree, data)) for x in data)

        # step 2: try removing neighbors to minimize entropy
        for n in list(cluster):
            if n == v:
                continue  # don't remove our seed, obviously
            new_c = cluster.copy()
            new_c.remove(n)
            new_e = dict((x, __entropy_w(new_c, x, degree, data)) for x in data[n])
            # if removing the neighbor decreases new entropy (for the node and
            # all its neighbors), then do so
            if sum(new_e.values()) < sum(entropies[x] for x in data[n]):
                cluster = new_c
                entropies.update(new_e)

        # boundary candidates
        c = (
            functools.reduce(lambda a, b: a | b, (set(data[x]) for x in cluster))
            - cluster
        )
        while c:
            n = c.pop()
            new_c = cluster.copy()
            new_c.add(n)
            new_e = dict((x, __entropy_w(new_c, x, degree, data)) for x in data[n])
            if sum(new_e.values()) < sum(entropies[x] for x in data[n]):
                cluster = new_c
                entropies.update(new_e)
                c &= set(data[n]) - cluster

        # remove the elements of the cluster from our candidate set; add cluster
        candidates -= cluster

        clusters.append(list(cluster))
    return clusters



# -*- coding: utf-8 -*-
"""
Created on Wed Jun 20 12:52:55 2018
@author: Vaiva & Tim
"""

from collections import defaultdict
import itertools
import numpy as np
import random
import itertools
import math
import networkx as nx
import scipy.sparse as sparse

__authors__ = ["Vaiva Vasiliauskaite", "T.S. Evans"]
__all__ = ["matrix_node_recursive_antichain_partition"]


def is_weakly_connected(graph, source_nodes, target_nodes):
    """
    Tests whether a list of source nodes in a graph have a path in either direction between a list of target nodes.

    Parameters
    ----------
    graph = networkx graph
    source_nodes = list of source nodes for paths
    target_nodes = list of target nodes for paths

    Returns
    -------
    Bool:
        True if there is a path from at least one source node to at least one target node or the other way round
        False otherwise
    """
    for s, t in itertools.product(source_nodes, target_nodes):
        if nx.has_path(graph, s, t) or nx.has_path(graph, t, s):
            return True
    return False


class Quality_matrix:
    """
    Quality measures for use in antichains. Similarity matrix implementation.
    """

    # class variables
    def __init__(self, node_id_dict, similarity_matrix, Lambda, with_replacement):
        # Initial Quality Measures
        self.similarity_matrix = similarity_matrix
        self.node_id_dict = node_id_dict
        self.strength = similarity_matrix.sum(axis=0)
        self.strength = {
            n: self.strength[0, node_id_dict[n]] for n in node_id_dict.keys()
        }  # sum over rows?
        self.total_weight = similarity_matrix.sum()  # /2
        self.Lambda = Lambda
        self.with_replacement = with_replacement

    def delta_strength_quality_unnormalised(self, partition1, partition2):
        """
        Using in-strength null model calculate the change in unnormalised quality if two partitions are combined.

        Definition is that used for weighted graph.

        Q = \sum_{u \in partition1} \sum_{v \in partition2}
            ( S_ij
             - k_i*k_j/W )
        where W = total strength of edges in the graph ((sum_{i,j}S_ij)/2),
        S_{ij} - i,j^th entry in the similarity matrix. For instance, A.A^T is successors-based similarity;
                A^T.A is predecessors-based similarity.

        Note this is not normalised.

        Note no test for connectedness of nodes in partitions.

        Note both partitions must be non-empty otherwise TypeError raised.

        Note no test to see if partitions are sets or if they share common elements.

        Input
        partition1 - iterable list or set of the nodes in first partition
        partition2 - iterable list or set of the nodes in second partition

        Return
        Contribution of the quality Q from the all pairs of nodes with one from partition1, second from partition2
        """

        return sum(
            [
                self.similarity_matrix[
                    self.node_id_dict[node1], self.node_id_dict[node2]
                ]
                - self.Lambda
                * self.strength[node1]
                * self.strength[node2]
                / self.total_weight
                for node1, node2 in itertools.product(partition1, partition2)
            ]
        )

    def quality_one_partition(self, partition):
        return sum(
            [
                self.similarity_matrix[
                    self.node_id_dict[node1], self.node_id_dict[node2]
                ]
                - self.Lambda
                * self.strength[node1]
                * self.strength[node2]
                / self.total_weight
                for node1, node2 in itertools.combinations(partition, 2)
            ]
        )

    def total_strength_quality_unnormalised(self, partitions):
        """
        Calculate the total unnormalised quality using strength null model

        Definition is that used for weighted graph.

        Q = \sum_{u \in partition1} \sum_{v \in partition2}
            ( S_ij
             - k_i*k_j/W )
        where W = total strength of edges in the graph ((sum_{i,j}S_ij)/2),
        S_{ij} - i,j^th entry in the similarity matrix. For instance, A.A^T is successors-based similarity;
                A^T.A is predecessors-based similarity.

        Note this is not normalised.

        Note no test for connectedness of nodes in partitions.

        Input
        -----
        partition - iterable list of sets of nodes in partitions so that
                    partition[p] is an set (or any iterable list) of the nodes in partition p

        Return
        ------
        Total value of the quality Q from the all pairs of nodes
        """
        return sum(
            [
                sum(
                    [
                        self.similarity_matrix[
                            self.node_id_dict[node1], self.node_id_dict[node2]
                        ]
                        - self.Lambda
                        * self.strength[node1]
                        * self.strength[node2]
                        / self.total_weight
                        for node1, node2 in itertools.combinations(p, 2)
                    ]
                )
                for p in partitions
            ]
        )


def get_edge_weight(G, node1, node2, weight_attribute="weight"):
    """
    Get Edge Weight

    Returns edge weight for edge in G from node1 to node2
    If edge exists and has weight_attribute, this value is returned.
    If edge exists but has not weight_attribute, 1 is returned.
    Otherwise 0 is returned

    Input
    -----
    G - networkx graph
    node1 - source node
    node2 - target node
    weight_attribute='weight' - attribute of edge containing weight value

    Return
    ------
    edge weight, 1 if edge exists but no weight attribute exists, 0 otherwise.

    """
    edge_data = G.get_edge_data(node1, node2)
    if edge_data is None:
        return 0
    elif weight_attribute in edge_data:
        return edge_data[weight_attribute]
    else:
        return 1


def get_node_attribute_value(G, node1, node_attribute=None):
    """
    Get Node Attribute Value

    Returns node attribute as a float.
    Otherwise 0.0 is returned

    Input
    -----
    G - networkx graph
    node1 - node
    node_attribute=None - attribute of node required

    Return
    ------
    node attribute as a float

    """
    try:
        node_data = G.node[node1][node_attribute]
        return float(node_data)
    except:
        pass
    return 0.0


def is_weakly_connected(graph, source_nodes, target_nodes):
    """
    Tests whether a list of source nodes in a graph have a path in either direction between a list of target nodes.

    Parameters
    ----------
    graph = networkx graph
    source_nodes = list of source nodes for paths
    target_nodes = list of target nodes for paths

    Returns
    -------
    Bool:
        True if there is a path from at least one source node to at least one target node or the other way round
        False otherwise
    """
    for s, t in itertools.product(source_nodes, target_nodes):
        if nx.has_path(graph, s, t) or nx.has_path(graph, t, s):
            return True
    return False


def coarse_grain(
    G,
    node_to_partition_label,
    partition_label_to_nodes,
    weight_attribute="weight",
    time_label="t",
    space_label="x",
):
    """
    Coarse Grain

    The new graph H has the partitions of G as the nodes in H.
    An edges from partition1 to partition2 in H is present if
    there is an edge from a node in G in partition1 of G to
    a node in G in partition2.  The total weight of the edge
    from partition1 to partition2 in H will be the sum of all
    the weights of all such edges in G
    from nodes in partition1 to nodes in partition2.
    If unweighted, weights are assumed to be 1.
    If time_label or space_label are set, these are assumed to be numerical
    values (e.g. coordinates) and nodes in the new graph get the average value from
    the partition of nodes they represent in the old graph.


    Input
    ----
    G - networkx graph
    node_to_partition_label - dictionary from G node key to its partition label
    partition_label_to_nodes - dictionary from partition label to the set of nodes in G in that partition
    weight_attribute='weight' - attribute on edge containing edge weight data
    time_label='t': Node key for time coordinate (used as y/vertical coordinate)
    space_label='x': Node key for space coordinate (used as x/horizontal coordinate)

    Return
    ------
    H - coarse grained graph, nodes are the partitions labels, weights are under eight_attribute of edges

    """
    H = nx.DiGraph()
    H.add_nodes_from(list(partition_label_to_nodes.keys()))
    for partition in partition_label_to_nodes.keys():
        # nodes_in_partition = partition_label_to_nodes[partition]
        number_in_partition = len(partition_label_to_nodes[partition])
        if time_label is not None:
            average_time = (
                sum(
                    [
                        get_node_attribute_value(G, n, node_attribute=time_label)
                        for n in partition_label_to_nodes[partition]
                    ]
                )
                / number_in_partition
            )
        H.nodes[partition][time_label] = average_time
        if space_label is not None:
            average_space = (
                sum(
                    [
                        get_node_attribute_value(G, n, node_attribute=space_label)
                        for n in partition_label_to_nodes[partition]
                    ]
                )
                / number_in_partition
            )
        H.nodes[partition][space_label] = average_space

    for partition1, partition2 in itertools.combinations(
        partition_label_to_nodes.keys(), 2
    ):
        w = sum(
            [
                get_edge_weight(G, node1, node2, weight_attribute)
                for node1, node2 in itertools.product(
                    partition_label_to_nodes[partition1],
                    partition_label_to_nodes[partition2],
                )
            ]
        )
        if w > 0:
            H.add_edge(partition1, partition2, weight_attribute=w)
        w = sum(
            [
                get_edge_weight(G, node2, node1, weight_attribute)
                for node1, node2 in itertools.product(
                    partition_label_to_nodes[partition1],
                    partition_label_to_nodes[partition2],
                )
            ]
        )
        if w > 0:
            H.add_edge(partition2, partition1, weight_attribute=w)
    return H


def similarity_matrix(DAG, similarity="intersection", neighbours="successors"):
    """
    Function to produce a similarity matrix based on neighbourhoods of nodes in DAG.

    Input
    -----
    DAG - networkx directed acyclic graph
    similarity - type of similarity of sets. Currently only implemented for the size of intersection
    neighbours - type of neighbours to consider in the similarity. Can be either successors or predecessors

    Return
    -----
    A - symmetric similarity matrix where entry A[i,j] represents similarity between nodes of indices i, j
    nodedict - dictionary of node names and their indices in the similarity matrix
    """

    nodes = list(DAG.nodes())
    nodedict = {}
    for i in range(len(nodes)):
        nodedict[nodes[i]] = i
    nodelist = list(nodedict.keys())

    A = (nx.adjacency_matrix(DAG, nodelist)).todense()

    if similarity == "intersection" and neighbours == "successors":
        A = A.dot(A.transpose())
        np.fill_diagonal(A, 0)
        return A, nodedict

    elif similarity == "intersection" and neighbours == "predecessors":
        A = (A.transpose()).dot(A)
        np.fill_diagonal(A, 0)
        return A, nodedict


def similarity_matrix_sparse(
    DAG, similarity="intersection", neighbours="successors", with_replacement=False
):
    """
    Function to produce a sparse similarity matrix based on neighbourhoods of nodes in DAG.

    Input
    -----
    DAG - networkx directed acyclic graph
    similarity - type of similarity of sets. Currently only implemented for the size of intersection
    neighbours - type of neighbours to consider in the similarity. Can be either successors or predecessors or both

    Return
    -----
    A - scipy sparse symmetric similarity matrix where entry A[i,j] represents similarity between nodes of indices i, j
    nodedict - dictionary of node names and their indices in the similarity matrix
    """

    nodes = list(DAG.nodes())
    nodedict = {}
    for i in range(len(nodes)):
        nodedict[nodes[i]] = i
    nodelist = list(nodedict.keys())

    A = nx.adjacency_matrix(DAG, nodelist)

    if similarity == "intersection" and neighbours == "successors":
        A = A.dot(A.transpose())
        if not with_replacement:
            A.setdiag(0)
        return A, nodedict

    elif similarity == "intersection" and neighbours == "predecessors":
        A = A.transpose().dot(A)
        if not with_replacement:
            A.setdiag(0)
        return A, nodedict
    elif similarity == "intersection" and neighbours == "both":
        A = A.transpose().dot(A) + A.dot(A.transpose())
        if not with_replacement:
            A.setdiag(0)
        return A, nodedict


def has_path_matrix(DAG, nodedict, cutoff=350):
    nodes = list(nodedict.keys())
    A = (nx.adjacency_matrix(DAG, nodes)).todense()
    A_sum = np.copy(A)
    if nx.is_directed_acyclic_graph(DAG):
        L_max = len(nx.dag_longest_path(DAG))
    else:
        L_max = cutoff
    current_length = 1
    while current_length <= L_max:
        A_sum = np.dot(A_sum, A)
        current_length += 1

    return (A_sum > 0).astype(np.int8)


def find_paths_sparse(A, length_max=10):
    """
    Tim's numpy path implementation updated by Vaiva to sparse matrices
    Scipy sparse matrix implementation to find all paths
    Given adjacency matrix A will find all the paths between all vertices

    Input
    A: numpy square adjacency matrix,can be weighted
    Return
    #tuple current_length,path_length,path_bool where
    #current_length = one more than the maximum length found.
                   If equals length_max then may have terminated because reachd maximum requested length
    #non_zero_entries = number of non-zero entries in P=(A)^current_length
    #path_length = matrix of longest paths lengths
                   path_length[target,source]= longest path from source to target
    path_bool = matrix of booleans indicating if path exists. Paths from vertex to slef (length zero) gives True on diagonal
                   path_bool[target,source]= True (False) if path from source to target
    """

    # Assume vertices start connected to selves only path path of length zero.
    m, n = np.shape(A)
    path_bool = sparse.eye(m, n, dtype=bool)
    path_bool = path_bool.tolil()
    path_length = sparse.lil_matrix((m, n), dtype="int32")
    current_length = 1
    P = A.copy()
    non_zero_entries = P.count_nonzero()

    while non_zero_entries > 0 and current_length < length_max:
        non_zero_entries = P.nonzero()
        path_bool[non_zero_entries[0], non_zero_entries[1]] = True
        path_length[non_zero_entries] = current_length
        P = P.dot(A)
        current_length += 1
        non_zero_entries = P.count_nonzero()
    return path_bool


def is_weakly_connected_matrix(path_matrix, nodedict, source_nodes, target_nodes):
    """
    Function to check whether nodes in the source_nodes are not weakly connected to
    nodes in the target_nodes.

    Input
    -----
    path_matrix - 1/0 matrix where entry P[i,j] = 1 if nodes with indices i,j are weakly connected
    nodedict - dictionary where keys are node names and values are their corresponding indices in the path matrix
    source_nodes - list of nodes
    target_nodes - list of nodes

    Return
    ------
    True - if nodes in source_nodes and target_nodes form a weakly_connected subgraph
    False - if not
    """
    source_nodes_id, target_nodes_id = [nodedict[s] for s in source_nodes], [
        nodedict[t] for t in target_nodes
    ]

    for s, t in itertools.product(source_nodes_id, target_nodes_id):
        if path_matrix[s, t] == 1 or path_matrix[t, s] == 1:
            return True
    return False


def node_matrix_greedy_antichain_partition(
    G,
    level,
    Lambda,
    with_replacement,
    random_on=False,
    seed=None,
    max_number_sweeps=None,
    backwards_forwards_on=True,
    forwards_backwards_on=False,
    Q_check_on=True,
    weight_attribute="weight",
):
    """
    In this implementation we iterate over nodes in the graph moving individual nodes until no changes occur .

    We start with each node in its own partition.
    In one sweep we look at each partition ac in turn.
    We find all the backwards-forwards neighbours of the nodes in partition ac
    and collect all their partition labels, excluding the current partition ac.
    For each of these we find if we can increase the quality function by merging
    current partition ac with one of its bf-neighbouring partitions. If we do we then
    do the merge removing the current partition ac.
    We continue the sweep looking at remaining partitions and trying to merge them.  Note that
    later partitions will see some previous partitions already merged.  The option to randomise the order
    that we visit the partitions will lead to different results.

    After each sweep, if at least one partition was merged then we sweep through again.
    We only stop when no more merges are found on one sweep, or if the number of sweeps exceed
    the maximum requested.

    Note that we do NOT create an explicit hybrid graph. We only use a weakly connected check
    of the

    Input
    -----
    G - networkx graph.
    random_on=False - if true will shuffle the order in which partitions are examined
    seed=None - used as seed if shuffle is on.  If None then time is used as seed
    max_number_sweeps=None - this is the maximum number of sweeps to consider. If less than 1 or None, then uses number of nodes.
    backwards_forwards_on=True  - find possible new partitions by making a backwards step then a forwards step from node being considered for a move
    forwards_backwards_on=False - find possible new partitions by making a forwards step then a backwards step from node being considered for a move
    Q_check_on=False - check to see if change in Q is correct by printing out total value and changes
    weight_attribute - edge attribute of weight. if None, unweighted quality functions are used. Note, weight must be integer

    Return
    ------
    tuple node_to_partition_label, partition_label_to_nodes
    where
    node_to_partition_label is a dictionary from node key to its partition label
    partition_label_to_nodes is a dictionary from partition label to the set of nodes in that partition

    """
    if not (forwards_backwards_on or backwards_forwards_on):
        raise ValueError(
            "At least one of forwards_backwards_on or backwards_forwards_on parameters must be True"
        )

    if backwards_forwards_on == True and forwards_backwards_on == True:
        adj_matrix, nodedict = similarity_matrix_sparse(
            G,
            similarity="intersection",
            neighbours="both",
            with_replacement=with_replacement,
        )

    elif backwards_forwards_on == True and forwards_backwards_on == False:
        # use in-degree quality with backwards-forwards step (predecessors)
        adj_matrix, nodedict = similarity_matrix_sparse(
            G,
            similarity="intersection",
            neighbours="predecessors",
            with_replacement=with_replacement,
        )
    elif forwards_backwards_on == True and backwards_forwards_on == False:
        # use out-degree quality with forwards-backwards step (successors)

        adj_matrix, nodedict = similarity_matrix_sparse(
            G,
            similarity="intersection",
            neighbours="successors",
            with_replacement=with_replacement,
        )
    Q = Quality_matrix(nodedict, adj_matrix, Lambda, with_replacement)
    Q_method = Q.delta_strength_quality_unnormalised
    Q_total = Q.total_strength_quality_unnormalised

    number_of_nodes = G.number_of_nodes()

    if max_number_sweeps is None or max_number_sweeps < 1:
        max_number_sweeps = number_of_nodes
    if random_on:
        random.seed(seed)
    # set up partition node dictionaries
    # These play the role of induced graphs
    # first place each node into its own partition
    node_to_partition_label = {}
    partition_label_to_nodes = {}
    next_partition_label = 0

    for n in G:
        node_to_partition_label[n] = next_partition_label
        partition_label_to_nodes[next_partition_label] = {n}
        next_partition_label += 1
    moved = True
    number_sweeps = 0

    if Q_check_on:
        Q_total_current = Q_total(partition_label_to_nodes.values())
    count = 0

    while moved == True and number_sweeps < max_number_sweeps:
        # Start of one sweep through all current partitions.
        # Check every partition ac in turn, doing the best merge you can for each partition ac.
        # Note the partition ac under study will be removed in the merge.
        # That means the list of partition labels will be altered and therefore
        # this list can not be used as a list of partition labels in the ac loop.
        # For that reason we need a deep copy of the current list of partition labels
        # which we get from the keys of the partition label to node set dictionary
        # Conveniently we can shuffle the list used in teh ac loop if we want to randomise the greedy
        # algorithm
        number_sweeps += 1
        number_moves = 0
        node_list = list(G.nodes())
        if random_on:
            random.shuffle(node_list)
        for n in node_list:
            count += 1
            # check to see if node n should be moved
            moved = False
            # ac is the partition containing node n
            ac = node_to_partition_label[n]
            # now find the contribution from Q that comes if we move n into its own partition
            partition_ac_no_n = set(partition_label_to_nodes[ac])
            partition_ac_no_n.discard(n)
            if len(partition_ac_no_n) > 0:
                delta_Q_remove_n = Q_method([n], partition_ac_no_n)
            else:
                delta_Q_remove_n = 0

            # now find the neighbouring partitions via backwards-forward step
            bf_nearest_neighbours_all = set()
            if backwards_forwards_on:
                for p in G.predecessors(n):
                    bf_nearest_neighbours_all.update(G.successors(p))
            if forwards_backwards_on:
                for p in G.successors(n):
                    bf_nearest_neighbours_all.update(G.predecessors(p))
            bf_nearest_neighbour_partitions = set(
                node_to_partition_label[bf_nn] for bf_nn in bf_nearest_neighbours_all
            )
            # remove current partition ac from neighbours
            try:
                bf_nearest_neighbour_partitions.remove(ac)
            except KeyError:  # ac is not in set, must have no neighbours or no b-f nearest neighbours
                pass

                # dictionary from partition label to delta quality value,
            # so delta_Q_nn[ac_nn] is change in quality if node n was to join partition ac_nn
            delta_Q_nn = {}
            # Loop round bf nearest neighbour partitions ac_nn.
            # Check each ac_nn partition to make sure it is not weakly connected to partition ac
            # then calculate the modularity change if partitions ac and ac_nn are merged
            for ac_nn in bf_nearest_neighbour_partitions:
                if not is_weakly_connected(G, [n], partition_label_to_nodes[ac_nn]):
                    delta_Q_nn[ac_nn] = Q_method([n], partition_label_to_nodes[ac_nn])
            if len(delta_Q_nn) > 0:
                # Note nice use of operator.itemgetter to get key with largest value
                # https://stackoverflow.com/questions/268272/getting-key-with-maximum-value-in-dictionary
                ac_max = max(delta_Q_nn, key=delta_Q_nn.get)
                if delta_Q_nn[ac_max] > delta_Q_remove_n and ac_max != ac:
                    # now merge partition ac into partition ac_max
                    number_moves += 1
                    node_to_partition_label[n] = ac_max
                    partition_label_to_nodes[ac_max].add(n)
                    partition_label_to_nodes[ac].remove(n)
                    if (
                        len(partition_label_to_nodes[ac]) == 0
                    ):  # no more elements in this partition
                        partition_label_to_nodes.pop(
                            ac, None
                        )  # remove ac from this dictionary

                    if Q_check_on:
                        dQ = delta_Q_nn[ac_max] - delta_Q_remove_n
                        Q_total_old = Q_total_current
                        Q_total_current = Q_total(partition_label_to_nodes.values())

                        _error = Q_total_current - Q_total_old - dQ
                    moved = True

                elif delta_Q_remove_n < 0:
                    number_moves += 1
                    node_to_partition_label[n] = next_partition_label
                    partition_label_to_nodes[next_partition_label] = {n}
                    next_partition_label += 1

        # keeping looping through all partitions until can not merge any more
    # keep doing new  sweeps as long as something changed
    return node_to_partition_label, partition_label_to_nodes


def matrix_node_recursive_antichain_partition(
    G,
    time_label="t",
    space_label="x",
    random_on=False,
    seed=None,
    max_number_sweeps=None,
    backwards_forwards_on=True,
    forwards_backwards_on=False,
    Q_check_on=True,
    plot_on=False,
    filenameroot=None,
    extlist=["pdf"],
    ScreenOn=False,
    Lambda=1,
    with_replacement=False,
):
    """

    Use , **kwargs in func defn and call with kawargs the dictionary for named arguments
    used for the partition

    """

    result_list = list()

    _matrix_node_recursive_antichain_partition_step(
        G,
        time_label=time_label,
        space_label=space_label,
        level=0,
        result_list=result_list,
        random_on=random_on,
        seed=seed,
        max_number_sweeps=max_number_sweeps,
        backwards_forwards_on=backwards_forwards_on,
        forwards_backwards_on=forwards_backwards_on,
        Q_check_on=Q_check_on,
        plot_on=plot_on,
        filenameroot=filenameroot,
        extlist=extlist,
        ScreenOn=ScreenOn,
        Lambda=Lambda,
        with_replacement=with_replacement,
    )

    return result_list


def _matrix_node_recursive_antichain_partition_step(
    G,
    Lambda,
    with_replacement,
    time_label="t",
    space_label="x",
    level=0,
    result_list=None,
    random_on=False,
    seed=None,
    max_number_sweeps=None,
    backwards_forwards_on=True,
    forwards_backwards_on=False,
    Q_check_on=True,
    plot_on=False,
    filenameroot=None,
    extlist=["pdf"],
    ScreenOn=False,
):
    # Internal routine to perform recursive version of node greedy
    result_list.append(None)
    (
        node_to_partition_label,
        partition_label_to_nodes,
    ) = node_matrix_greedy_antichain_partition(
        G,
        random_on=random_on,
        seed=seed,
        max_number_sweeps=max_number_sweeps,
        backwards_forwards_on=backwards_forwards_on,
        forwards_backwards_on=forwards_backwards_on,
        Q_check_on=Q_check_on,
        level=level,
        Lambda=Lambda,
        with_replacement=with_replacement,
    )
    if len(partition_label_to_nodes.keys()) == G.number_of_nodes():
        return
        # optional plot

    new_G = coarse_grain(G, node_to_partition_label, partition_label_to_nodes)
    _matrix_node_recursive_antichain_partition_step(
        new_G,
        time_label=time_label,
        space_label=space_label,
        level=level + 1,
        result_list=result_list,
        random_on=random_on,
        seed=seed,
        max_number_sweeps=max_number_sweeps,
        backwards_forwards_on=backwards_forwards_on,
        forwards_backwards_on=forwards_backwards_on,
        Q_check_on=Q_check_on,
        plot_on=plot_on,
        filenameroot=filenameroot,
        extlist=extlist,
        ScreenOn=ScreenOn,
        Lambda=Lambda,
        with_replacement=with_replacement,
    )
    result_list[level] = {
        "level": level,
        "n_to_p": node_to_partition_label,
        "p_to_n": partition_label_to_nodes,
    }
    return



from contextlib import contextmanager

try:
    import igraph as ig
except ModuleNotFoundError:
    ig = None

try:
    import graph_tool as gt
except ModuleNotFoundError:
    gt = None

import networkx as nx
import sys
import os
import numpy as np


@contextmanager
def suppress_stdout():
    """
    Suppress the standard out messages.
    """
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = devnull
        sys.stderr = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr


def __from_nx_to_graph_tool(g: object, directed: bool = None) -> object:
    """

    :param g:
    :param directed:
    :return:
    """

    if directed is None:
        directed = g.is_directed()

    if gt is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install igraph to use the selected feature."
        )

    gt_g = gt.Graph(directed=directed)

    node_map = {v: i for i, v in enumerate(g.nodes())}

    gt_g.add_vertex(len(node_map))
    gt_g.add_edge_list([(node_map[u], node_map[v]) for u, v in g.edges()])

    return gt_g, {v: k for k, v in node_map.items()}


def __from_graph_tool_to_nx(
    graph: object, node_map: dict = None, directed: bool = None
) -> object:

    if directed is None:
        directed = graph.is_directed()

    if directed:
        tp = nx.DiGraph()
    else:
        tp = nx.Graph()

    tp.add_nodes_from([int(v) for v in graph.vertices()])
    tp.add_edges_from([(int(e.source()), int(e.target())) for e in graph.edges()])
    if node_map:
        nx.relabel_nodes(tp, node_map, copy=False)

    return tp


def __from_nx_to_igraph(g: object, directed: bool = None) -> ig.Graph:
    """
    :param g:
    :param directed:
    :return:
    """
    from networkx.algorithms import bipartite

    if ig is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install igraph to use the selected feature."
        )

    if directed is None:
        directed = g.is_directed()

    gi = ig.Graph(directed=directed)

    a_r = {}
    if bipartite.is_bipartite(g):
        A, B = bipartite.sets(g)
        for a in A:
            a_r[a] = 0
        for b in B:
            a_r[b] = 1

    ## Two problems to handle:
    # 1)in igraph, names have to be str.
    # 2)since we can ask to compute metrics with found communities and the the original graph, we need to keep
    # the original nodes types in communities. Therefore we need to handle some transparent conversion for non-str nodes
    if type(list(g.nodes)[0]) is str:  # if nodes are string, no problem
        gi.add_vertices([n for n in g.nodes()])
        gi.add_edges([(u, v) for (u, v) in g.edges()])

    else:
        if set(range(len(g.nodes))) == set(
            g.nodes()
        ):  # if original names are well formed contiguous ints, keep this for efficiency.
            # Put these int as str with identitiers in the name attribute
            gi.add_vertices(len(g.nodes))
            gi.add_edges([(u, v) for (u, v) in g.edges()])
            gi.vs["name"] = ["\\" + str(n) for n in g.nodes()]
        else:  # if names are not well formed ints, convert to string and use the identifier to remember
            # converting back to int
            # convert = {str(x):x for x in g.nodes()}
            gi.add_vertices(["\\" + str(n) for n in g.nodes()])
            gi.add_edges([("\\" + str(u), "\\" + str(v)) for (u, v) in g.edges()])

    if bipartite.is_bipartite(g):
        gi.vs["type"] = [
            a_r[name] if type(name) == int else a_r[int(name.replace("\\", ""))]
            for name in gi.vs["name"]
        ]

    edgelist = nx.to_pandas_edgelist(g)
    for attr in edgelist.columns[2:]:
        gi.es[attr] = edgelist[attr]

    return gi


def __from_igraph_to_nx(gi: ig.Graph, directed: bool = None) -> object:
    """

    :param gi:
    :param directed:
    :return:
    """

    if gi is None:
        raise ModuleNotFoundError(
            "Optional dependency not satisfied: install igraph to use the selected feature."
        )

    if directed is None:
        directed = gi.is_directed()

    if directed:
        tp = nx.DiGraph()
    else:
        tp = nx.Graph()

    for e in gi.es:
        tp.add_edge(gi.vs[e.source]["name"], gi.vs[e.target]["name"], **e.attributes())

    return tp


def convert_graph_formats(
    graph: object, desired_format: object, directed: bool = None
) -> object:
    """Converts from/to networkx/igraph


    :param graph: original graph object
    :param desired_format: desired final type. Either nx.Graph or ig.Graph
    :param directed: boolean, default **False**
    :return: the converted graph
    :raises TypeError: if input graph is neither an instance of nx.Graph nor ig.Graph
    """
    if isinstance(graph, desired_format):
        return graph
    elif desired_format is nx.Graph:
        return __from_igraph_to_nx(graph, directed)
    elif ig is not None and desired_format is ig.Graph:
        return __from_nx_to_igraph(graph, directed)
    else:
        raise TypeError(
            "The graph object should be either a networkx or an igraph one."
        )


def nx_node_integer_mapping(graph: object) -> tuple:
    """Maps node labels from strings to integers.

    :param graph: networkx graph
    :return: if the node labels are string: networkx graph, dictionary <numeric_id, original_node_label>, false otherwise
    """

    convert = False
    for nid in graph.nodes():
        if isinstance(nid, str):
            convert = True
            break

    if convert:
        node_map = {}
        label_map = {}
        if isinstance(graph, nx.Graph):
            for nid, name in enumerate(graph.nodes()):
                node_map[nid] = name
                label_map[name] = nid

            nx.relabel_nodes(graph, label_map, copy=False)
            return graph, node_map
        else:
            raise ValueError("graph must be a networkx Graph object")

    return graph, None


def remap_node_communities(communities: object, node_map: dict) -> list:
    """Apply a map to the obtained communities to retreive the original node labels

    :param communities: NodeClustering object
    :param node_map: dictionary <numeric_id, node_label>
    :return: remapped communities
    """

    cms = []
    for community in communities:
        community = [node_map[n] for n in community]
        cms.append(community)
    communities = cms
    return communities


def affiliations2nodesets(affiliations: dict) -> dict:
    """
    Transform community format to nodesets

    Representation expected in input: dictionary, key=node, value=list/set of snapshot_affiliations ID
    Representation in output: bidict, key=community ID , value=set of nodes

    :param affiliations:bidict, key=community ID , value=set of nodes
    :return: dict, key=community ID , value=set of nodes
    """

    if affiliations is None:
        return None

    asNodeSets = dict()

    if len(affiliations) == 0:
        return asNodeSets

    for n, coms in affiliations.items():
        if isinstance(coms, str) or isinstance(coms, int) or isinstance(coms, np.int32):
            coms = [coms]
        for c in coms:
            asNodeSets.setdefault(c, set())
            asNodeSets[c].add(n)

    return asNodeSets



import numpy as np
import networkx as nx
from collections import defaultdict

"""
 Reference implementation 
 https://github.com/nidhisridhar/Fuzzy-Community-Detection
"""


def __reachable(i, theta_cores, fuzz_d, visited):
    # Returns indices of cores(in theta_cores) that are reachable from theta_cores[ i ]
    reach = []
    flag = True
    index = -1
    num_cores = len(theta_cores)
    while flag:
        if index == len(reach):
            flag = False
        if index == -1:
            flag = False
            for j in range(num_cores):
                if visited[j] == 0 and i != j:
                    if fuzz_d[theta_cores[i]][theta_cores[j]] > 0:
                        visited[j] = 1
                        reach.append(j)
                        flag = True
        else:
            for j in range(num_cores):
                if visited[j] == 0 and index != j:
                    if fuzz_d[theta_cores[index]][theta_cores[j]] > 0:
                        visited[j] = 1
                        reach.append(j)
                        flag = True
        index += 1
    return np.array(reach)


def __gran_embed(core, c, fuzz_d):
    # Return Normalized Granular Embeddedness of theta-core with community C
    num = 0
    den = 0
    c = np.array(c)
    c = np.append(c, core)
    n = len(fuzz_d[0])
    for i in range(n):
        num += np.min(fuzz_d[c, i])
        den += np.max(fuzz_d[c, i])
    return float(num) / den


def fuzzy_comm(graph, theta, eps, r):
    """
    Takes adjacency_mat(n*n) , theta , eps (epsilon) , and radius(r)
    and returns an n*c matrix where c is the number of communities and the
    i,jth value is the membership of node i in community j

    :param graph: networkx graph
    :param theta:
    :param eps:
    :param r:
    :return:
    """

    adjacency_mat = nx.to_numpy_matrix(graph)

    theta_cores = []
    num_vertices = adjacency_mat.shape[0]

    # Fuzzy granule initialization
    # gran = [i for i in range(num_vertices)]

    # Calculate distance between all vertices
    dist = list(nx.all_pairs_shortest_path_length(graph))

    # Membership values between all nodes
    fuzz_d = np.zeros(shape=adjacency_mat.shape).astype(float)
    for i in range(num_vertices):
        nid, n_dist = dist[i]
        for j in graph.nodes():
            if j in n_dist and n_dist[j] <= r:
                fuzz_d[nid][j] = 1 / float(1 + n_dist[j])
    _sum = np.sum(fuzz_d, axis=1)

    # Normalization of Membership
    for i in range(num_vertices):
        fuzz_d[i] = fuzz_d[i] / float(_sum[i])

        # Theta-cores Finding
    for i in range(num_vertices):
        if np.sum(fuzz_d[:, i]) >= theta:
            theta_cores.append(i)
    theta_cores = np.array(theta_cores)
    num_cores = len(theta_cores)
    _sum = np.sum(fuzz_d[:, theta_cores], axis=1)
    k = 0
    for i in range(num_vertices):
        fuzz_d[i] = fuzz_d[i] / _sum[k]
        k += 1

    # Finding Fuzzy Communities
    communities = []
    visited = np.zeros(num_cores)

    for i in range(num_cores):
        if visited[i] == 0:
            c = [theta_cores[i]]
            visited[i] = 1
            reach = __reachable(i, theta_cores, fuzz_d, visited.copy())
            for core_ind in reach:
                if __gran_embed(theta_cores[core_ind], c, fuzz_d) > eps:
                    c.append(theta_cores[core_ind])
                    visited[core_ind] = 1
            communities.append(c)

    cms = []
    for c in communities:
        cms.append([int(n) for n in c])

    # fuzzy association to communities
    fuzz_assoc = defaultdict(dict)
    for i in range(num_vertices):
        for j in range(len(cms)):
            fuzz_assoc[i][j] = np.sum(fuzz_d[i, cms[j]])

    return cms, fuzz_assoc



from cdlib.benchmark.internal import rdyn
from cdlib import TemporalClustering, NamedClustering
from collections import defaultdict
import dynetx as dn

__all__ = ["RDyn"]


def RDyn(
    size: int = 300,
    iterations: int = 5,
    avg_deg: int = 15,
    sigma: float = 0.6,
    lambdad: float = 1,
    alpha: float = 2.5,
    paction: float = 1,
    prenewal: float = 0.8,
    quality_threshold: float = 0.5,
    new_node: float = 0.0,
    del_node: float = 0.0,
    max_evts: int = 1,
    simplified: bool = True,
) -> [dn.DynGraph, object]:
    """
    RDyn is a syntetic dynamic network generator with time-dependent ground-truth partitions having tunable quality (in terms of conductance).
    Communities' ids are aligned across time and a predefined number of merge/plit events are planted in between consecutive stable iterations.

    :param size: Number of nodes
    :param iterations: Number of stable iterations
    :param avg_deg: Average node degree. Int, default 15
    :param sigma: Percentage of node's edges within a community. Float, default .6
    :param lambdad: Community size distribution exponent. Float, default 1
    :param alpha:  Degree distribution exponent. Float, default 2.5
    :param paction: Probability of node action. Float, default 1
    :param prenewal: Probability of edge renewal. Float, default, .8
    :param quality_threshold: Conductance quality threshold for stable iteration. Float, default .5
    :param new_node: Probability of node appearance. Float, default 0
    :param del_node: Probability of node vanishing. Float, default 0
    :param max_evts: Max number of community events for stable iteration. Int, default 1
    :param simplified: Simplified execution. Boolean, default True. (NB: when True an approximation of the original process is executed - some network characteristics can deviate from the expected ones)

    :return: A dynetx DynGraph, the TemporalClustering object

    :Example:

    >>> from cdlib.benchmark import RDyn
    >>> G, coms = RDyn(n=300)

    :References:

    Rossetti, Giulio. "RDyn: graph benchmark handling community dynamics." Journal of Complex Networks 5.6 (2017): 893-912.

    .. note:: Reference implementation: https://github.com/GiulioRossetti/RDyn
    """
    tc = TemporalClustering()
    dg = dn.DynGraph()
    rdb = rdyn.RDynV2(
        size=size,
        iterations=iterations,
        avg_deg=avg_deg,
        quality_threshold=quality_threshold,
        sigma=sigma,
        lambdad=lambdad,
        alpha=alpha,
        paction=paction,
        prenewal=prenewal,
        new_node=new_node,
        del_node=del_node,
        max_evts=max_evts,
    )

    t = 0
    for g, communities in rdb.execute(simplified=simplified):
        dg.add_interactions_from(g.edges(), t=t, e=t)
        sg = dg.time_slice(t)
        nc = NamedClustering(
            communities,
            sg,
            "RDyn",
            {
                "size": size,
                "iterations": iterations,
                "avg_deg": avg_deg,
                "quality_threshold": quality_threshold,
                "sigma": sigma,
                "lambdad": lambdad,
                "alpha": alpha,
                "paction": paction,
                "prenewal": prenewal,
                "new_node": new_node,
                "del_node": del_node,
                "max_evts": max_evts,
                "simplified": simplified,
            },
            overlap=False,
        )
        tc.add_clustering(nc, time=t)
        t += 1

    return dg, tc



import warnings
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import networkx as nx
from cdlib import NodeClustering
from cdlib.utils import convert_graph_formats
from community import community_louvain

__all__ = ["plot_network_clusters", "plot_community_graph"]

# [r, b, g, c, m, y, k, 0.8, 0.2, 0.6, 0.4, 0.7, 0.3, 0.9, 0.1, 0.5]
COLOR = (
    (1, 0, 0),
    (0, 0, 1),
    (0, 0.5, 0),
    (0, 0.75, 0.75),
    (0.75, 0, 0.75),
    (0.75, 0.75, 0),
    (0, 0, 0),
    (0.8, 0.8, 0.8),
    (0.2, 0.2, 0.2),
    (0.6, 0.6, 0.6),
    (0.4, 0.4, 0.4),
    (0.7, 0.7, 0.7),
    (0.3, 0.3, 0.3),
    (0.9, 0.9, 0.9),
    (0.1, 0.1, 0.1),
    (0.5, 0.5, 0.5),
)


def __filter(partition: list, top_k: int, min_size: int) -> list:
    if isinstance(min_size, int) and min_size > 0:
        partition = list(filter(lambda nodes: len(nodes) >= min_size, partition))
    if isinstance(top_k, int) and top_k > 0:
        partition = partition[:top_k]
    return partition


def plot_network_clusters(
    graph: object,
    partition: NodeClustering,
    position: dict = None,
    figsize: tuple = (8, 8),
    node_size: int = 200,
    plot_overlaps: bool = False,
    plot_labels: bool = False,
    cmap: object = None,
    top_k: int = None,
    min_size: int = None,
) -> object:
    """
    Plot a graph with node color coding for communities.

    :param graph: NetworkX/igraph graph
    :param partition: NodeClustering object
    :param position: A dictionary with nodes as keys and positions as values. Example: networkx.fruchterman_reingold_layout(G). By default, uses nx.spring_layout(g)
    :param figsize: the figure size; it is a pair of float, default (8, 8)
    :param node_size: int, default 200
    :param plot_overlaps: bool, default False. Flag to control if multiple algorithms memberships are plotted.
    :param plot_labels: bool, default False. Flag to control if node labels are plotted.
    :param cmap: str or Matplotlib colormap, Colormap(Matplotlib colormap) for mapping intensities of nodes. If set to None, original colormap is used.
    :param top_k: int, Show the top K influential communities. If set to zero or negative value indicates all.
    :param min_size: int, Exclude communities below the specified minimum size.

    Example:

    >>> from cdlib import algorithms, viz
    >>> import networkx as nx
    >>> g = nx.karate_club_graph()
    >>> coms = algorithms.louvain(g)
    >>> pos = nx.spring_layout(g)
    >>> viz.plot_network_clusters(g, coms, pos)
    """
    if not isinstance(cmap, (type(None), str, matplotlib.colors.Colormap)):
        raise TypeError(
            "The 'cmap' argument must be NoneType, str or matplotlib.colors.Colormap, "
            "not %s." % (type(cmap).__name__)
        )

    partition = __filter(partition.communities, top_k, min_size)
    graph = convert_graph_formats(graph, nx.Graph)
    if position is None:
        position = nx.spring_layout(graph)

    n_communities = len(partition)
    if n_communities == 0:
        warnings.warn("There are no communities that match the filter criteria.")
        return None

    if cmap is None:
        n_communities = min(n_communities, len(COLOR))
        cmap = matplotlib.colors.ListedColormap(COLOR[:n_communities])
    else:
        cmap = plt.cm.get_cmap(cmap, n_communities)
    _norm = matplotlib.colors.Normalize(vmin=0, vmax=n_communities - 1)
    fontcolors = list(
        map(
            lambda rgb: ".15" if np.dot(rgb, [0.2126, 0.7152, 0.0722]) > 0.408 else "w",
            [cmap(_norm(i))[:3] for i in range(n_communities)],
        )
    )

    plt.figure(figsize=figsize)
    plt.axis("off")

    filtered_nodelist = list(np.concatenate(partition))
    filtered_edgelist = list(
        filter(
            lambda edge: len(np.intersect1d(edge, filtered_nodelist)) == 2,
            graph.edges(),
        )
    )
    fig = nx.draw_networkx_nodes(
        graph, position, node_size=node_size, node_color="w", nodelist=filtered_nodelist
    )
    fig.set_edgecolor("k")
    nx.draw_networkx_edges(graph, position, alpha=0.5, edgelist=filtered_edgelist)
    if plot_labels:
        nx.draw_networkx_labels(
            graph,
            position,
            font_color=".8",
            labels={node: str(node) for node in filtered_nodelist},
        )
    for i in range(n_communities):
        if len(partition[i]) > 0:
            if plot_overlaps:
                size = (n_communities - i) * node_size
            else:
                size = node_size
            fig = nx.draw_networkx_nodes(
                graph,
                position,
                node_size=size,
                nodelist=partition[i],
                node_color=[cmap(_norm(i))],
            )
            fig.set_edgecolor("k")
    if plot_labels:
        for i in range(n_communities):
            if len(partition[i]) > 0:
                nx.draw_networkx_labels(
                    graph,
                    position,
                    font_color=fontcolors[i],
                    labels={node: str(node) for node in partition[i]},
                )

    return fig


def plot_community_graph(
    graph: object,
    partition: NodeClustering,
    figsize: tuple = (8, 8),
    node_size: int = 200,
    plot_overlaps: bool = False,
    plot_labels: bool = False,
    cmap: object = None,
    top_k: int = None,
    min_size: int = None,
) -> object:
    """
    Plot a algorithms-graph with node color coding for communities.

    :param graph: NetworkX/igraph graph
    :param partition: NodeClustering object
    :param figsize: the figure size; it is a pair of float, default (8, 8)
    :param node_size: int, default 200
    :param plot_overlaps: bool, default False. Flag to control if multiple algorithms memberships are plotted.
    :param plot_labels: bool, default False. Flag to control if node labels are plotted.
    :param cmap: str or Matplotlib colormap, Colormap(Matplotlib colormap) for mapping intensities of nodes. If set to None, original colormap is used..
    :param top_k: int, Show the top K influential communities. If set to zero or negative value indicates all.
    :param min_size: int, Exclude communities below the specified minimum size.

    Example:

    >>> from cdlib import algorithms, viz
    >>> import networkx as nx
    >>> g = nx.karate_club_graph()
    >>> coms = algorithms.louvain(g)
    >>> viz.plot_community_graph(g, coms)
    """

    cms = __filter(partition.communities, top_k, min_size)

    node_to_com = {}
    for cid, com in enumerate(cms):
        for node in com:
            if node not in node_to_com:
                node_to_com[node] = cid
            else:
                # duplicating overlapped node
                alias = "%s_%s" % (node, cid)
                node_to_com[alias] = cid
                edges = [(alias, y) for y in graph.neighbors(node)]
                graph.add_edges_from(edges)

    # handling partial coverage
    s = nx.subgraph(graph, node_to_com.keys())

    # algorithms graph construction
    c_graph = community_louvain.induced_graph(node_to_com, s)
    node_cms = [[node] for node in c_graph.nodes()]

    return plot_network_clusters(
        c_graph,
        NodeClustering(node_cms, None, ""),
        nx.spring_layout(c_graph),
        figsize=figsize,
        node_size=node_size,
        plot_overlaps=plot_overlaps,
        plot_labels=plot_labels,
        cmap=cmap,
    )



from cdlib import NodeClustering, FuzzyNodeClustering, EdgeClustering
import json
import gzip

__all__ = [
    "write_community_csv",
    "read_community_csv",
    "write_community_json",
    "read_community_json",
    "read_community_from_json_string",
]


def write_community_csv(
    communities: object, path: str, delimiter: str = ",", compress: bool = False
):
    """
    Save community structure to comma separated value (csv) file.

    :param communities: a NodeClustering object
    :param path: output filename
    :param delimiter: column delimiter
    :param compress: wheter to copress the csv, default False

    :Example:

    >>> import networkx as nx
    >>> from cdlib import algorithms, readwrite
    >>> g = nx.karate_club_graph()
    >>> coms = algorithms.louvain(g)
    >>> readwrite.write_community_csv(coms, "communities.csv", ",")

    """
    if compress:
        op = gzip.open
    else:
        op = open

    with op(path, "wt") as f:
        for cid, community in enumerate(communities.communities):
            res = delimiter.join(list(map(str, community)))
            f.write("%s\n" % res)


def read_community_csv(
    path: str, delimiter: str = ",", nodetype: type = str, compress: bool = False
) -> object:
    """
    Read community list from comma separated value (csv) file.

    :param path: input filename
    :param delimiter: column delimiter
    :param nodetype: specify the type of node labels, default str
    :param compress: wheter the file is compressed or not, default False
    :return: NodeClustering object

    :Example:

    >>> import networkx as nx
    >>> from cdlib import algorithms, readwrite
    >>> g = nx.karate_club_graph()
    >>> coms = algorithms.louvain(g)
    >>> readwrite.write_community_csv(coms, "communities.csv", ",")
    >>> coms = readwrite.read_community_csv(coms, "communities.csv", ",", str)

    """
    communities = []

    if compress:
        op = gzip.open
    else:
        op = open

    with op(path, "rt") as f:
        for row in f:
            community = list(map(nodetype, row.rstrip().split(delimiter)))
            communities.append(list(community))

    return NodeClustering(communities, None, "")


def write_community_json(communities: object, path: str, compress: bool = False):
    """
    Generate a JSON representation of the clustering object

    :param communities: a cdlib clustering object
    :param path: output filename
    :param compress: wheter to copress the JSON, default False
    :return: a JSON formatted string representing the object

    :Example:

    >>> import networkx as nx
    >>> from cdlib import algorithms, readwrite
    >>> g = nx.karate_club_graph()
    >>> coms = algorithms.louvain(g)
    >>> readwrite.write_community_json(coms, "communities.json")
    """

    partition = {
        "communities": communities.communities,
        "algorithm": communities.method_name,
        "params": communities.method_parameters,
        "overlap": communities.overlap,
        "coverage": communities.node_coverage,
    }

    try:
        partition["allocation_matrix"] = communities.allocation_matrix
    except AttributeError:
        pass

    js_dmp = json.dumps(partition)

    if compress:
        op = gzip.open
    else:
        op = open

    with op(path, "wt") as f:
        f.write(js_dmp)


def read_community_json(path: str, compress: bool = False) -> object:
    """
    Read community list from JSON file.

    :param path: input filename
    :param compress: wheter the file is in a copress format, default False
    :return: a Clustering object

    :Example:

    >>> import networkx as nx
    >>> from cdlib import algorithms, readwrite
    >>> g = nx.karate_club_graph()
    >>> coms = algorithms.louvain(g)
    >>> readwrite.write_community_json(coms, "communities.json")
    >>> readwrite.read_community_json(coms, "communities.json")
    """

    if compress:
        op = gzip.open
    else:
        op = open

    with op(path, "rt") as f:
        coms = json.load(f)

    nc = NodeClustering(
        [list(c) for c in coms["communities"]],
        None,
        coms["algorithm"],
        coms["params"],
        coms["overlap"],
    )
    nc.node_coverage = coms["coverage"]

    if "allocation_matrix" in coms:
        nc.__class__ = FuzzyNodeClustering
        nc.allocation_matrix = coms["allocation_matrix"]

    if type(nc.communities[0][0]) is list:
        cms = []
        for c in nc.communities:
            cm = []
            for e in c:
                cm.append(tuple(e))
            cms.append(list(cm))
        nc.communities = cms
        nc.__class__ = EdgeClustering

    return nc


def read_community_from_json_string(json_repr: str) -> object:
    """
    Read community list from JSON file.

    :param json_repr: json community representation
    :return: a Clustering object

    :Example:

    >>> import networkx as nx
    >>> from cdlib import algorithms, readwrite
    >>> g = nx.karate_club_graph()
    >>> coms = algorithms.louvain(g)
    >>> readwrite.write_community_json(coms, "communities.json")
    >>> with open("community.json") as f:
    >>>     cr = f.read()
    >>>     readwrite.write_community_from_json_string(cr)
    """

    coms = json.loads(json_repr)

    nc = NodeClustering(
        [list(c) for c in coms["communities"]],
        None,
        coms["algorithm"],
        coms["params"],
        coms["overlap"],
    )
    nc.node_coverage = coms["coverage"]

    if "allocation_matrix" in coms:
        nc.__class__ = FuzzyNodeClustering
        nc.allocation_matrix = coms["allocation_matrix"]

    if type(nc.communities[0][0]) is list:
        cms = []
        for c in nc.communities:
            cm = []
            for e in c:
                cm.append(tuple(e))
            cms.append(tuple(cm))
        nc.communities = cms
        nc.__class__ = EdgeClustering

    return nc



import Eva

from collections import defaultdict
from cdlib import AttrNodeClustering

import networkx as nx

from cdlib.utils import convert_graph_formats

from cdlib.algorithms.internal.ILouvain import ML2

__all__ = ["eva", "ilouvain"]


def eva(
    g_original: object,
    labels: dict,
    weight: str = "weight",
    resolution: float = 1.0,
    alpha: float = 0.5,
) -> AttrNodeClustering:

    """
    The Eva algorithm extends the Louvain approach in order to deal with the attributes of the nodes (aka Louvain Extended to Vertex Attributes).
    It optimizes - combining them linearly - two quality functions, a structural and a clustering one, namely Newman's modularity and purity, estimated as the product of the frequencies of the most frequent labels carried by the nodes within the communities.
    A parameter alpha tunes the importance of the two functions: an high value of alpha favors the clustering criterion instead of the structural one.


    **Supported Graph Types**

    ========== ======== ======== ======== ==============
    Undirected Directed Weighted Temporal Node Attribute
    ========== ======== ======== ======== ==============
    Yes        No       No       No       Yes
    ========== ======== ======== ======== ==============

    :param g_original: a networkx/igraph object
    :param labels: dictionary specifying for each node (key) a dict (value) specifying the name attribute (key) and its value (value)
    :param weight: str, optional the key in graph to use as weight. Default to 'weight'
    :param resolution: double, optional  Will change the size of the communities, default to 1.
    :param alpha: float, assumed in [0,1], optional Will tune the importance of modularity and purity criteria, default to 0.5
    :return: AttrNodeClustering object

    :Example:

    >>> from cdlib.algorithms import eva
    >>> import networkx as nx
    >>> import random
    >>> l1 = ['A', 'B', 'C', 'D']
    >>> l2 = ["E", "F", "G"]
    >>> g_attr = nx.barabasi_albert_graph(100, 5)
    >>> labels=dict()
    >>> for node in g_attr.nodes():
    >>>    labels[node]={"l1":random.choice(l1), "l2":random.choice(l2)}
    >>> communities = eva(g_attr, labels, alpha=0.8)

    :References:

    Citraro, S., & Rossetti, G. (2019, December). Eva: Attribute-Aware Network Segmentation. In International Conference on Complex Networks and Their Applications (pp. 141-151). Springer, Cham.

    .. note:: Reference implementation: https://github.com/GiulioRossetti/Eva/tree/master/Eva
    """

    g = convert_graph_formats(g_original, nx.Graph)
    nx.set_node_attributes(g, labels)

    coms, coms_labels = Eva.eva_best_partition(
        g, weight=weight, resolution=resolution, alpha=alpha
    )

    # Reshaping the results
    coms_to_node = defaultdict(list)
    for n, c in coms.items():
        coms_to_node[c].append(n)

    coms_eva = [list(c) for c in coms_to_node.values()]
    return AttrNodeClustering(
        coms_eva,
        g_original,
        "Eva",
        coms_labels,
        method_parameters={
            "weight": weight,
            "resolution": resolution,
            "alpha": alpha,
        },
    )


def ilouvain(g_original: object, labels: dict) -> AttrNodeClustering:
    """
    The I-Louvain algorithm extends the Louvain approach in order to deal only with the scalar attributes of the nodes.
    It optimizes Newman's modularity combined with an entropy measure.


    **Supported Graph Types**

    ========== ======== ======== ======== ==============
    Undirected Directed Weighted Temporal Node Attribute
    ========== ======== ======== ======== ==============
    Yes        No       No       No       Yes
    ========== ======== ======== ======== ==============

    :param g_original: a networkx/igraph object
    :param labels: dictionary specifying for each node (key) a dict (value) specifying the name attribute (key) and its value (value)
    :return: AttrNodeClustering object

    :Example:

    >>> from cdlib.algorithms import ilouvain
    >>> import networkx as nx
    >>> import random
    >>> l1 = [0.1, 0.4, 0.5]
    >>> l2 = [34, 3, 112]
    >>> g_attr = nx.barabasi_albert_graph(100, 5)
    >>> labels=dict()
    >>> for node in g_attr.nodes():
    >>>    labels[node]={"l1":random.choice(l1), "l2":random.choice(l2)}
    >>> id = dict()
    >>> for n in g.nodes():
    >>>     id[n] = n
    >>> communities = ilouvain(g_attr, labels, id)

    :References:

    Combe D., Largeron C., Géry M., Egyed-Zsigmond E. "I-Louvain: An Attributed Graph Clustering Method". <https://link.springer.com/chapter/10.1007/978-3-319-24465-5_16> In: Fromont E., De Bie T., van Leeuwen M. (eds) Advances in Intelligent Data Analysis XIV. IDA (2015). Lecture Notes in Computer Science, vol 9385. Springer, Cham

    """

    g = convert_graph_formats(g_original, nx.Graph)
    nx.set_node_attributes(g, labels)
    nid = dict()
    for n in g.nodes():
        nid[n] = n

    algo = ML2(g, labels, nid)
    coms = algo.findPartition()

    # Reshaping the results
    coms_to_node = defaultdict(list)
    for n, c in coms.items():
        coms_to_node[c].append(n)

    coms_ilouv = [list(c) for c in coms_to_node.values()]

    return AttrNodeClustering(coms_ilouv, g_original, "ILouvain")



import numpy as np
from sklearn.cluster import KMeans
import sklearn.preprocessing as pre
import networkx as nx
from collections import defaultdict


def Kcut(data, kmax):
    """
    此算法执行的是Kcut算法  Ruan J , Zhang W . An Efficient Spectral Algorithm for Network Community Discovery and
    Its Applications to Biological and Social Networks[C]// Data Mining, 2007. ICDM 2007. Seventh IEEE International Conference on. IEEE, 2007.
    data:图的邻接矩阵  kmax:k-menas算法k的最大值
    """
    Q = -100.0  # 模块度值
    # 计算社团的所有边
    c = 0.0  # 所有的边数
    for i in range(data.shape[0]):
        for j in range(data.shape[0]):
            if data[i, j] == 1:
                c = c + 1
    c = c / 2
    for k in range(2, 2 * kmax + 1):
        Qk = 0  # 当前社团的模块度
        clusters = []  # 当前的聚类结果,是一个n*1列向量，每一行代表一个样本的类标签
        # 计算图的度
        W = np.zeros((data.shape[0], data.shape[0]))
        D = np.zeros([data.shape[0], data.shape[0]])
        # 计算对角度矩阵
        for i in range(data.shape[0]):
            for j in range(data.shape[0]):
                W[i, j] = np.linalg.norm(data[i, :] - data[i, :], 2)
            value0 = sum(W[i, :]) / data.shape[0]
            for j in range(data.shape[0]):
                if W[i, j] < value0:
                    W[i, j] = 0
        for i in range(data.shape[0]):
            D[i, i] = np.sum(W[i, :])
        # 计算拉普拉斯矩阵
        L = D - W
        value, vector = np.linalg.eig(L)  # 对拉普拉斯矩阵进行特征值分解
        seq = np.argsort(value)  # 计算特征值从小到大排列的序号
        seq = seq[0:k]
        vector = vector[:, seq]
        vector = np.array(vector)
        for i in range(vector.shape[0]):  # 将向量的单位长度进行单位化
            vector[i, :] = vector[i, :] / (np.linalg.norm(vector[i, :], 2) + 0.00001)
        estimator = KMeans(n_clusters=k)  # k-means聚类
        vector = np.real(vector)
        vector = pre.minmax_scale(vector)
        estimator.fit(vector)
        clusters = estimator.labels_  # 获得最终的聚类标签,类标签为一个行向量
        """计算社团划分结果的模块度"""
        for i in range(k + 1):
            # 获得第i社团的节点
            e = 0  # 两个顶点均在第i个社团
            a = 0  # 至少一个顶点在第i个社团
            node = np.argwhere(clusters == i)
            node = node[:, 0]
            node = np.array(node)
            for m in range(data.shape[0]):
                for n in range(data.shape[0]):
                    if data[m, n] == 1:
                        temp1 = []
                        temp2 = []
                        temp1 = np.argwhere(node == m)
                        temp2 = np.argwhere(node == n)
                        if len(temp1) != 0:
                            a = a + 1
                            if len(temp2) != 0:
                                e = e + 1
            a = a / 2
            e = e / 2
            Qk = Qk + e / c - (a / c) * (a / c)
        if Qk > Q:  # 当前的模块度优于之前的模块度
            Q = Qk
            result = clusters  # 保存社团划分结果

    coms = defaultdict(list)
    for n, c in enumerate(result):
        coms[c].append(n)

    return list(coms.values())


def kcut_exec(g, kmax):
    nodes = list(g.nodes())
    data = nx.to_numpy_array(g, nodelist=nodes)
    result = Kcut(data, kmax)

    communities = []
    for c in result:
        cms = []
        for n in c:
            cms.append(nodes[n])
        communities.append(cms)

    return communities



import numpy as np


def __normalize(adjacency_matrix):
    adjacency_matrix = adjacency_matrix / np.sum(adjacency_matrix, axis=0)
    return adjacency_matrix


def __expand(a, m):
    return np.linalg.matrix_power(a, m)


def __inflate(a, r):
    return __normalize(np.power(a, r))


def __get_clusters(a):
    clusters = []
    for i, r in enumerate((a > 0).tolist()):
        if r[i]:
            clusters.append(a[i, :] > 0)
    clustering_map = {}
    for cn, c in enumerate(clusters):
        for x in [i for i, x in enumerate(c) if x]:
            clustering_map[cn] = clustering_map.get(cn, []) + [x]
    return clustering_map


def __create_adj_matrix(initial_matrix, first_column, second_column):
    for i in range(first_column.size):
        x = int(first_column[i])
        y = int(second_column[i])
        # connecting edges, as it is undirected graph edge is set both ways.
        initial_matrix[x][y] = 1.0
        initial_matrix[y][x] = 1.0
        # Adding self loops
        initial_matrix[x][x] = 1.0
        initial_matrix[y][y] = 1.0
    # removing the extra row and column
    initial_matrix = np.delete(initial_matrix, 0, 0)
    initial_matrix = np.delete(initial_matrix, 0, 1)
    return initial_matrix


def buildCluMap(clusters):
    custMap = {}
    for key in clusters:
        listele = clusters.get(key)
        for ele in listele:
            custMap[ele] = key
    return custMap


def markov(graph, max_loop=1000):
    """
    An Efficient Algorithm for Large-scale Detection of Protein Families (Nucleic Acids Research 2002)
    Anton Enright, Stijn Van Dongen, and Christos Ouzounis

    https://github.com/HarshHarwani/markov-clustering-for-graphs

    :param graph:
    :param max_loop:
    :return:
    """
    edges = np.array(list(map(np.float64, graph.edges())))

    # taking the max from the array to create new matrix of required dimension
    max_value = int(np.amax(edges))

    # extracting two columns from the original two dimensional array
    first_column = edges[:, 0]
    second_column = edges[:, 1]

    # creating the initial empty array, creating one extra dimension to start the index from 1.
    inital_matrix = np.ndarray(shape=(max_value + 1, max_value + 1))

    # creating the adjacencyMatrix
    adjacency_matrix = __create_adj_matrix(inital_matrix, first_column, second_column)

    cls = {}

    for m in range(2, 7):
        for r in np.arange(1.1, 2.2, 0.2):

            # normalizing the matrix
            matrix = __normalize(adjacency_matrix)

            # configuration parameters
            for i in range(max_loop):
                # maintaining a previous copy for checking convergence
                prev = matrix.copy()
                matrix = __expand(matrix, m)
                matrix = __inflate(matrix, r)

                # convergence condition
                if np.array_equal(matrix, prev):
                    break

            clusters = __get_clusters(matrix)
            cls = clusters

    communities = []

    for part in cls.values():
        com = []
        for eid in part:
            com.extend(list(map(int, edges[eid])))
        communities.append(list(set(com)))

    return communities



import random
import collections

"""
Newman, Mark EJ, and Elizabeth A. Leicht. 
"Mixture algorithms and exploratory analysis in networks." 
Proceedings of the National Academy of Sciences 104.23 (2007): 9564-9569.
"""


class EM_nx(object):
    def __init__(self, g, k, max_iter=100):
        self.g = g
        self.n = len(self.g.nodes)
        self.k = k
        self.pi = []
        self.theta = []
        self.max_iter = max_iter

    def e_step(self, q):
        for i in range(self.n):
            q.append([])
            norm = 0.0
            for g in range(self.k):
                x = self.pi[g]
                for j in self.g.neighbors(i):
                    x *= self.theta[g][j]
                q[i].append(x)
                norm += x
            for g in range(self.k):
                q[i][g] /= norm

    def m_step(self, q):
        for g in range(self.k):
            sum1 = 0.0
            sum3 = 0.0
            for i in range(self.n):
                sum1 += q[i][g]
                sum2 = 0.0
                for j in self.g.neighbors(i):
                    sum2 += q[j][g]
                self.theta[g][i] = sum2  # update theta
                sum3 += q[i][g] * len(list(self.g.neighbors(i)))
            self.pi[g] = sum1 / self.n  # update pi
            for i in range(self.n):
                self.theta[g][i] /= sum3  # norm

    def execute(self):
        # initial parameters
        X = [1.0 + random.random() for i in range(self.k)]
        norm = sum(X)
        self.pi = [x / norm for x in X]

        for i in range(self.k):
            Y = [1.0 + random.random() for j in range(self.n)]
            norm = sum(Y)
            self.theta.append([y / norm for y in Y])

        q_old = []
        for iter_time in range(self.max_iter):
            q = []
            # E-step
            self.e_step(q)
            # M-step
            self.m_step(q)

            if iter_time != 0:
                deltasq = 0.0
                for i in range(self.n):
                    for g in range(self.k):
                        deltasq += (q_old[i][g] - q[i][g]) ** 2
                # print "delta: ", deltasq
                if deltasq < 0.05:
                    # print "iter_time: ", iter_time
                    break

            q_old = []
            for i in range(self.n):
                q_old.append([])
                for g in range(self.k):
                    q_old[i].append(q[i][g])

        communities = collections.defaultdict(lambda: set())
        for i in range(self.n):
            c_id = 0
            cur_max = q[i][0]
            for j in range(1, self.k):
                if q[i][j] > cur_max:
                    cur_max = q[i][j]
                    c_id = j
            communities[c_id].add(i)
        return list(communities.values())



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 11 11:05:52 2021
@author: Gianmarco
References: Desmond J.Highama Gabriela Kalnaa Milla Kibble, Spectral clustering and its use in bioinformatics
            Journal of Computational and Applied Mathematics
"""
import networkx as nx
import numpy as np

# Use this function to calculate a diagonal matrix useful to build  Laplacian matrix
def __diagonal_matrix(W):
    D = np.zeros([W.shape[0], W.shape[1]])
    for i in range(W.shape[0]):
        s = 0
        for j in range(W.shape[1]):
            s = s + W[i, j]

        D[i, i] = s
    return D


def spectral_communities(g, kmax, projection_on_smaller_class=True, scaler=None):
    """
    Constructor
    :param g: a networkx Graph object
    :param kmax: maximum number of desired communities
    :param projection_on_smaller_class: a boolean value that if True then it project a bipartite network in the smallest class of node. (default is True)
    :param scaler: the function to scale the fielder’s vector to apply KMeans
    """
    from networkx import bipartite
    from numpy import linalg as LA
    from scipy.linalg import eigh

    from sklearn.cluster import KMeans
    from sklearn.metrics import silhouette_score
    from sklearn.preprocessing import MinMaxScaler

    # conditional statement to build weight matrix in two different cases
    if not bipartite.is_bipartite(g):
        b = list(g.nodes())
        W = nx.adjacency_matrix(g, b)

    else:
        top_nodes = {n for n, d in g.nodes(data=True) if d["bipartite"] == 0}
        bottom_nodes = set(g) - top_nodes

        b = list(bottom_nodes)
        t = list(top_nodes)

        A = np.array(bipartite.biadjacency_matrix(g, t, b).todense())
        A_t = A.transpose()
        if projection_on_smaller_class:
            W = np.matmul(A_t, A)
        else:
            W = np.matmul(A, A_t)

    D = __diagonal_matrix(W)
    I = np.linalg.inv(D)
    L = LA.multi_dot([np.sqrt(I), (D - W), np.sqrt(I)])
    E, V = eigh(L)  # returns eigenvalue and eigenvector sorted by eigenvalue
    fielder_vec = np.matmul(
        np.sqrt(I), V[:, 1]
    )  # the fielder vector is the second eigenvector ov V

    x = [i for i in range(len(fielder_vec))]
    fielder_nodes = [(fielder_vec[i], x[i]) for i in range(len(x))]
    fielder_nodes.sort()

    y = [e[0] for e in fielder_nodes]

    X = [[y[i], x[i]] for i in range(len(x))]
    # normalizing X before applying kMeans
    X = scaler.fit_transform(X)

    # Compute silhouette score to choose the best number of communities
    sill = []
    K = range(1, kmax + 1)

    for k in K:
        km = KMeans(n_clusters=k)
        km = km.fit(X)
        if k != 1:
            sill.append(silhouette_score(X, km.labels_))

    k_best = 2 + max(range(len(sill)), key=sill.__getitem__)
    kmeans = KMeans(n_clusters=k_best)
    kmeans.fit(X)

    # build communities from clustering label
    communities = [[] for _ in range(k_best)]
    for i, e in enumerate(kmeans.labels_):
        communities[e].append(b[fielder_nodes[i][1]])

    return communities  # , b, fielder_vec



import scipy as sp
import math

logBase = 2


def partialEntropyAProba(proba: float) -> float:
    if proba == 0:
        return 0
    return -proba * math.log(proba, logBase)


def coverEntropy(cover: list, allNodes: list) -> float:
    allEntr = []
    for com in cover:
        fractionIn = len(com) / len(allNodes)
        allEntr.append(sp.stats.entropy([fractionIn, 1 - fractionIn], base=logBase))

    return sum(allEntr)


# https://github.com/RapidsAtHKUST/CommunityDetectionCodes
def comPairConditionalEntropy(cl: set, clKnown: set, allNodes: set) -> float:

    # cl1,cl2, snapshot_communities (set of nodes)
    # H(Xi|Yj ) =H(Xi, Yj ) − H(Yj )
    # h(a,n) + h(b,n) + h(c,n) + h(d,n)
    # −h(b + d, n)−h(a + c, n)
    # a: count agreeing on not belonging
    # b: count disagreeing : not in 1 but in 2
    # c: count disagreeing : not in 2 but in 1
    # d: count agreeing on belonging
    nbNodes = len(allNodes)

    a = len((allNodes - cl) - clKnown) / nbNodes
    b = len(clKnown - cl) / nbNodes
    c = len(cl - clKnown) / nbNodes
    d = len(cl & clKnown) / nbNodes

    if partialEntropyAProba(a) + partialEntropyAProba(d) > partialEntropyAProba(
        b
    ) + partialEntropyAProba(c):
        entropyKnown = sp.stats.entropy(
            [len(clKnown) / nbNodes, 1 - len(clKnown) / nbNodes], base=logBase
        )
        conditionalEntropy = sp.stats.entropy([a, b, c, d], base=logBase) - entropyKnown
    else:
        conditionalEntropy = sp.stats.entropy(
            [len(cl) / nbNodes, 1 - len(cl) / nbNodes], base=logBase
        )

    return conditionalEntropy  # *nbNodes


def coverConditionalEntropy(
    cover: list, coverRef: list, allNodes: set, normalized: bool = False
) -> float:
    # cover and coverRef and list of set

    allMatches = []

    for com in cover:
        matches = [
            (com2, comPairConditionalEntropy(com, com2, allNodes)) for com2 in coverRef
        ]
        bestMatch = min(matches, key=lambda c: c[1])
        HXY_part = bestMatch[1]
        if normalized:
            HX = partialEntropyAProba(len(com) / len(allNodes)) + partialEntropyAProba(
                (len(allNodes) - len(com)) / len(allNodes)
            )
            if HX == 0:
                HXY_part = 1
            else:
                HXY_part = HXY_part / HX
        allMatches.append(HXY_part)

    to_return = sum(allMatches)
    if normalized:
        to_return = to_return / len(cover)
    return to_return


def onmi(
    cover: list, coverRef: list, allNodes: set = None, variant: str = "LFK"
) -> float:
    """
    Compute Overlapping NMI

    This implementation allows to compute 3 versions of the overlapping NMI
    LFK: The original implementation proposed by Lacichinetti et al.(1). The normalization of mutual information is done community by community
    MGH: In (2), McDaid et al. argued that the original NMI normalization was flawed and introduced a new (global) normalization by the max of entropy
    MGH_LFK: This is a variant of the LFK method introduced in (2), with the same type of normalization but done globally instead of at each community

    Results are checked to be similar to the C++ implementations by the authors of (2): https://github.com/aaronmcdaid/Overlapping-NMI

    :param cover: set of set of nodes
    :param coverRef:set of set of nodes
    :param allNodes:
    :param variant:
    :param adjustForChance:
    :return:

    :Reference:

    1. Lancichinetti, A., Fortunato, S., & Kertesz, J. (2009). Detecting the overlapping and hierarchical community structure in complex networks. New Journal of Physics, 11(3), 033015.
    2. McDaid, A. F., Greene, D., & Hurley, N. (2011). Normalized mutual information to evaluate overlapping community finding algorithms. arXiv preprint arXiv:1110.2515. Chicago

    """
    if (len(cover) == 0 and len(coverRef) != 0) or (
        len(cover) != 0 and len(coverRef) == 0
    ):
        return 0
    if cover == coverRef:
        return 1

    if allNodes is None:
        allNodes = {n for c in coverRef for n in c}
        allNodes |= {n for c in cover for n in c}

    if variant == "LFK":
        HXY = coverConditionalEntropy(cover, coverRef, allNodes, normalized=True)
        HYX = coverConditionalEntropy(coverRef, cover, allNodes, normalized=True)
    else:
        HXY = coverConditionalEntropy(cover, coverRef, allNodes)
        HYX = coverConditionalEntropy(coverRef, cover, allNodes)

    HX = coverEntropy(cover, allNodes)
    HY = coverEntropy(coverRef, allNodes)

    NMI = -10
    if variant == "LFK":
        NMI = 1 - 0.5 * (HXY + HYX)
    elif variant == "MGH_LFK":
        NMI = 1 - 0.5 * (HXY / HX + HYX / HY)
    elif variant == "MGH":
        IXY = 0.5 * (HX - HXY + HY - HYX)
        NMI = IXY / (max(HX, HY))

    if math.isnan(NMI) or round(NMI, 2) < 0 or round(NMI, 2) > 1:
        print("NMI: %s  from %s %s %s %s " % (NMI, HXY, HYX, HX, HY))
        raise Exception("incorrect NMI")
    return NMI



from collections import Counter, defaultdict
import itertools

try:
    import igraph as ig
except ModuleNotFoundError:
    ig = None
import numpy as np
import operator

# import logging


#############################
# Fuzzy Modularity Measures #
#############################


def nepusz_modularity(G, cover):
    raise NotImplementedError("See the CONGA 2010 paper")


def zhang_modularity(G, cover):
    raise NotImplementedError(
        """See 'Identification of overlapping algorithms structure in
        complex networks using fuzzy C-means clustering'"""
    )


def nicosia_modularity(G, cover):
    raise NotImplementedError(
        """See 'Extending the definition of
        modularity to directed graphs with overlapping communities'"""
    )


#############################
# Crisp modularity measures #
#############################


def count_communities(G, cover):
    """
    Helper for lazar_modularity.
    Returns a dict {v:count} where v is a vertex id and
    count is the number of different communities it is
    assigned to.
    """
    counts = {i.index: 0 for i in G.vs}
    for community in cover:
        for v in community:
            counts[v] += 1
    return counts


def get_weights(G):
    """
    Given a graph G, returns a list of weights. If the graph is unweighted,
    returns a list of 1s the same length as the number of edges.
    """
    try:
        # asssumes weight as an attribute name means graph is weighted.
        weights = G.es["weight"]
    except KeyError:
        # unweighted means all weights are 1.
        weights = [1 for e in G.es]
    return weights


def get_single_lazar_modularity(G, community, weights, counts):
    """
    Returns the lazar modularity of a single algorithms.
    """
    totalInternalWeight = sum(weights[G.es[e].index] for e in community)  # m_c in paper
    numVerticesInCommunity = len(community)  # V_c in paper
    numPossibleInternalEdges = numVerticesInCommunity * (numVerticesInCommunity - 1) / 2
    if numPossibleInternalEdges == 0:
        return 0
    edgeDensity = (
        totalInternalWeight / numPossibleInternalEdges / numVerticesInCommunity
    )
    interVsIntra = 0
    comm = set(community)
    for v in community:
        interVsIntraInternal = 0
        neighbors = G.neighbors(v)
        degree = len(neighbors)  # k_i in paper
        numCommunitiesWithin = counts[v]  # s_i in paper
        for n in neighbors:
            weight = weights[G.get_eid(v, n)]
            if n in comm:
                interVsIntraInternal += weight
            else:
                interVsIntraInternal -= weight
        interVsIntraInternal /= degree * numCommunitiesWithin
        interVsIntra += interVsIntraInternal
    return edgeDensity * interVsIntra


def lazar_modularity(G, cover):
    """
    Returns the crisp modularity measure as defined by Lazar et al. 2009
    Defined as the average edge density times normalized difference
    between inter and intracommunity edges for each algorithms.
    See CONGA 2010 or Lazar's paper for a precise definition.
    """
    numCommunities = len(cover)  # |C| in the paper
    totalModularity = 0  # M_c in the paper
    weights = get_weights(G)
    counts = count_communities(G, cover)
    for c in cover:
        totalModularity += get_single_lazar_modularity(G, c, weights, counts)
    averageModularity = 1 / numCommunities * totalModularity  #  M in the paper
    return averageModularity


##################################
# Classes for overlapping covers #
##################################


class CrispOverlap(object):
    """
    TODO
    """

    def __init__(
        self,
        graph,
        covers,
        modularities=None,
        optimal_count=None,
        modularity_measure="lazar",
    ):
        """
        Initializes a CrispOverlap object with the given parameters.
            Graph: The graph to which the object refers
            covers: a dict of VertexCovers, also referring to this graph, of the form {k : v}
                where k is the number of clusters and v is the vertexCluste
            modularities (optional): a dict of modularities of the form {c:m} where c is
                the number of clusters and m is the modularity.
            optimal_count (optional): A hint for the number of clusters to use.
            modularity_measure (optional): The name of the modularity function to use.
                Right now, the only choice is "lazar."
        """
        # Possibly figure out a better data structure like a merge
        # list that contains all information needed?

        # So far only know of Lazar's measure for crisp overlapping.
        self._measureDict = {"lazar": lazar_modularity}
        self._covers = covers
        self._graph = graph
        self._optimal_count = optimal_count
        self._modularities = modularities
        if modularity_measure in self._measureDict:
            self._modularity_measure = modularity_measure
        else:
            raise KeyError("Modularity measure not found.")

    def __getitem__(self, numClusters):
        """
        Returns the cover with the given number of clusters.
        """
        if not numClusters:
            raise KeyError("Number of clusters must be a positive integer.")
        return self._covers[numClusters]

    def __iter__(self):
        """
        Iterates over the covers in the list.
        """
        return (v for v in list(self._covers.values()))

    def __len__(self):
        """
        Returns the number of covers in the list.
        """
        return len(self._covers)

    def __bool__(self):
        """
        Returns True when there is at least one cover in the list.
        """
        return bool(self._covers)

    def __str__(self):
        """
        Returns a string representation of the list of covers.
        """
        return "{0} vertices in {1} possible covers.".format(
            len(self._graph.vs), len(self._covers)
        )

    def as_cover(self):
        """
        Returns the optimal cover (by modularity) from the object.
        """
        return self._covers[self.optimal_count]

    def recalculate_modularities(self):
        """
        Recalculates the modularities and optimal count using the modularity_measure.
        """
        modDict = {}
        for cover in self._covers.values():
            modDict[len(cover)] = self._measureDict[self._modularity_measure](
                self._graph, cover
            )
        self._modularities = modDict
        self._optimal_count = max(
            iter(self._modularities.items()), key=operator.itemgetter(1)
        )[0]
        return self._modularities

    @property
    def modularities(self):
        """
        Returns the a dict {c : m} where c is the number of clusters
        in the cover and m is the modularity. If modularity has not
        been calculated, it recalculates it for all covers. Otherwise,
        it returns the stored dict.
        Note: Call recalculate_modularities to recalculate the modularity.
        """
        if self._modularities:
            return self._modularities
        self._modularities = self.recalculate_modularities()
        return self._modularities

    @property
    def optimal_count(self):
        """Returns the optimal number of clusters for this dendrogram.
        If an optimal count hint was given at construction time and
        recalculate_modularities has not been called, this property simply returns the
        hint. If such a count was not given, this method calculates the optimal cover
        by maximizing the modularity along all possible covers in the object.
        Note: Call recalculate_modularities to recalculate the optimal count.
        """
        if self._optimal_count is not None:
            return self._optimal_count
        else:
            modularities = self.modularities
            self._optimal_count = max(
                list(modularities.items()), key=operator.itemgetter(1)
            )[0]
            return self._optimal_count

    def pretty_print_cover(self, numClusters, label="CONGA_index"):
        """
        Takes a cover in vertex-id form and prints it nicely
        using label as each vertex's name.
        """
        cover = self._covers[numClusters]
        # if label == 'CONGA_index':
        pp = [self._graph.vs[num] for num in [cluster for cluster in cover]]
        # else:
        #    pp = [G.vs[num][label] for num in [cluster for cluster in cover]]
        for count, comm in enumerate(pp):
            print("Community {0}:".format(count))
            for v in comm:
                print("\t {0}".format(v.index if label == "CONGA_index" else v[label]))
            print()

    def make_fuzzy(self):
        """
        TODO. see CONGA 2010
        """
        pass


#


###################################################################################################################################################


# TODO:
#    * only call fix_betweennesses when needed


def congo(OG, h=2):
    """
    Provides an Implementation of the CONGO algorithm defined by Steve Gregory
    in his 2010 paper "A Fast Algorithm to Find Overlapping Communities in Networks."
    The parameters are OG, the graph on which the analysis is to be performed, and h,
    the length of the longest shortest path that Congo is to consider.
    """

    # logging.basicConfig(filename='congo.log',level=logging.DEBUG)
    G = OG.copy()

    # Just in case the original graph is disconnected
    if not G.is_connected():
        raise RuntimeError("Congo only makes sense for connected graphs.")

    # initializing attributes of copied graph
    G.vs["CONGA_orig"] = [i.index for i in OG.vs]
    G.es["eb"] = 0
    G.vs["pb"] = [
        {uw: 0 for uw in itertools.combinations(G.neighbors(vertex), 2)}
        for vertex in G.vs
    ]

    # initializing all pair and edge betweennesses
    do_initial_betweenness(G, h)
    nClusters = 1

    # The first cover is simply the entire connected graph.
    allCovers = {nClusters: ig.VertexCover(OG)}
    while G.es:

        # logging.info("%d edges remaining", len(G.es))
        # get the edge with the max edge betweenness, and its betweenness.
        maxEdge, maxEb = max(enumerate(G.es["eb"]), key=operator.itemgetter(1))
        G.vs["vb"] = G.betweenness(cutoff=h)

        # since split betweennes is upper bounded by vertex betweenness, we
        # only need to look at the vertices for which the vertex betweenness
        # is greater than the max edge betweenness. (We multiply by 2
        # because our edge betweenness calculations yield values in both
        # directions.)

        # TODO check if I need to multiply by 2
        vInteresting = [i for i, b in enumerate(G.vs["vb"]) if 2 * b > maxEb]

        # logging.info("Vertices to examine: %s", vInteresting)
        splitInstr = max_split_betweenness(G, vInteresting)

        # split if max split betweenness > max edge betweenness
        if splitInstr is None or splitInstr[0] <= maxEb:
            split = delete_edge(G, maxEdge, h)
        else:
            split = split_vertex(G, splitInstr[1], splitInstr[2], h)

        if split:
            # there must be a new algorithms
            comm = G.components().membership
            cover = get_cover(G, OG, comm)
            nClusters += 1
            # short circuit stuff would go here.
            allCovers[nClusters] = cover
    return CrispOverlap(OG, allCovers)


def delete_edge(G, edge, h):
    """
    Given a graph G and one of its edges in tuple form, checks if the deletion
    splits the graph.
    """

    tup = G.es[edge].tuple

    # logging.info("Deleted: %s", tup)

    neighborhood = get_neighborhood_edge(G, tup, h)
    # subtracts local betweennesses in the region, as discussed
    # in the paper
    do_local_betweenness(G, neighborhood, h, operator.neg)
    G.delete_edges(edge)
    fix_betweennesses(G)
    # adds back in local betweennesses after the deletion
    do_local_betweenness(G, neighborhood, h, operator.pos)
    return check_for_split(G, tup)


def fix_pair_betweennesses(G):
    """
    Given a graph G, makes sure that all of the pair betweennesses
    listed as attributes remain possible, and removes those that are not.
    Also adds new attributes where new edges have been added.
    """
    for v in G.vs:
        toDel = []
        neededPairs = {uw for uw in itertools.combinations(G.neighbors(v), 2)}
        for pair in v["pb"]:
            if pair not in neededPairs:
                toDel.append(pair)
        for d in toDel:
            del v["pb"][d]
        for pair in neededPairs:
            if pair not in v["pb"]:
                v["pb"][pair] = 0


def fix_edge_betweennesses(G):
    """
    Given a graph G, makes sure that every edge has a betweenness
    score assigned to it.
    """
    for e in G.es:
        if e["eb"] is None:
            e["eb"] = 0


def fix_betweennesses(G):
    """
    Fixes the pair and edge betweennesses such that every attribute is up to date.
    """
    fix_pair_betweennesses(G)
    fix_edge_betweennesses(G)


def split_vertex(G, vToSplit, instr, h):
    """
    Splits the vertex v into two new vertices, each with
    edges depending on s. Returns True if the split
    divided the graph, else False.
    """
    neighborhood = get_neighborhood_vertex(G, vToSplit, h)
    do_local_betweenness(G, neighborhood, h, operator.neg)
    new_index = G.vcount()
    G.add_vertex()
    G.vs[new_index]["CONGA_orig"] = G.vs[vToSplit]["CONGA_orig"]
    G.vs[new_index]["pb"] = {
        uw: 0 for uw in itertools.combinations(G.neighbors(vToSplit), 2)
    }

    # adding all relevant edges to new vertex, deleting from old one.
    toAdd = list(zip(itertools.repeat(new_index), instr[0]))
    toDelete = list(zip(itertools.repeat(vToSplit), instr[0]))
    G.add_edges(toAdd)
    G.delete_edges(toDelete)
    neighborhood.append(new_index)
    fix_betweennesses(G)
    # logging.info("split: %d, %s", vToSplit, instr)
    do_local_betweenness(G, neighborhood, h, operator.pos)
    # check if the two new vertices are disconnected.
    return check_for_split(G, (vToSplit, new_index))


def max_split_betweenness(G, vInteresting):
    """
    Performs the greedy algorithm discussed in the 2007 CONGA paper
    to approximate the maximum split betweenness. Returns a tuple
    (a, b, c) where a is the maximum score, b the vertex to split
    to acheive the score, and c a list of the instructions for which
    neighbors to connect to each side of the split.
    """
    maxSplitBetweenness = 0
    vToSplit = None
    # for every vertex of interest, we want to figure out the maximum score achievable
    # by splitting the vertices in various ways, and return that optimal split
    for v in vInteresting:
        clique = create_clique(G, v, G.vs["pb"][v])
        if clique.size < 4:
            continue

        # initialize a list on how we will map the neighbors to the collapsing matrix
        vMap = [[ve] for ve in G.neighbors(v)]

        # we want to keep collapsing the matrix until we have a 2x2 matrix and its
        # score. Then we want to remove index j from our vMap list and concatenate
        # it with the vMap[i]. This begins building a way of keeping track of how
        # we are splitting the vertex and its neighbors
        while clique.size > 4:
            i, j, clique = reduce_matrix(clique)
            vMap[i] += vMap.pop(j)

        if clique[0, 1] >= maxSplitBetweenness:
            maxSplitBetweenness = clique[0, 1]
            vToSplit = v
            splitInstructions = vMap
    if vToSplit is None:
        return None

    return maxSplitBetweenness, vToSplit, splitInstructions


def do_initial_betweenness(G, h):
    """
    Given a graph G and a depth h, calculates all edge and pair betweennesses
    and updates G's attributes to reflect the new scores.
    """
    # Not guaranteed to work on multigraphs.
    all_pairs_shortest_paths = []
    # Counter for normalizing scores
    pathCounts = Counter()
    for ver in G.vs:
        # logging.info("initializing betweennesses for %d", ver.index)
        neighborhood = get_neighborhood_vertex(G, ver, h)
        neighborhood.remove(ver.index)
        # for i, v in enumerate(neighborhood):
        s_s_shortest_paths = G.get_all_shortest_paths(ver, to=neighborhood)  # [i+1:])
        all_pairs_shortest_paths += s_s_shortest_paths

    # to ignore duplicate edges, uncomment the next line.
    # all_pairs_shortest_paths = set(tuple(p) for p in all_pairs_shortest_paths)
    for path in all_pairs_shortest_paths:
        pathCounts[(path[0], path[-1])] += 1

    # logging.info("updating all betweenness attributes...")
    for path in all_pairs_shortest_paths:
        if len(path) <= h + 1:
            update_betweenness(G, path, pathCounts[(path[0], path[-1])], operator.pos)


def do_local_betweenness(G, neighborhood, h, op=operator.pos):
    """
    Given a neighborhood and depth h, recalculates all betweennesses
    confined to the neighborhood. If op is operator.neg, it subtracts these
    betweennesses from the current ones. Otherwise, it adds them.
    """
    all_pairs_shortest_paths = []
    pathCounts = Counter()
    for i, v in enumerate(neighborhood):
        s_s_shortest_paths = G.get_all_shortest_paths(v, to=neighborhood)  # [i+1:])
        all_pairs_shortest_paths += s_s_shortest_paths
    neighSet = set(neighborhood)
    neighSize = len(neighborhood)
    apsp = []
    for path in all_pairs_shortest_paths:
        # path does not go out of region
        if len(neighSet | set(path)) == neighSize:
            pathCounts[(path[0], path[-1])] += 1  # can improve
            apsp.append(path)
    for path in apsp:
        if len(path) <= h + 1:
            update_betweenness(G, path, pathCounts[(path[0], path[-1])], op)


def update_betweenness(G, path, count, op):
    """
    Given a shortest path in G, along with a count of paths
    that length, to determine weight, updates the edge and
    pair betweenness dicts with the path's new information.
    """
    weight = op(1.0 / count)
    pos = 0
    while pos < len(path) - 2:
        G.vs[path[pos + 1]]["pb"][order_tuple((path[pos], path[pos + 2]))] += weight
        G.es[G.get_eid(path[pos], path[pos + 1])]["eb"] += weight
        pos += 1
    if pos < len(path) - 1:
        G.es[G.get_eid(path[pos], path[pos + 1])]["eb"] += weight


def get_cover(G, OG, comm):
    """
    Given the graph, the original graph, and a algorithms
    membership list, returns a vertex cover of the communities
    referring back to the original algorithms.
    """
    coverDict = defaultdict(list)
    for i, community in enumerate(comm):
        coverDict[community].append(int(G.vs[i]["CONGA_orig"]))
    return ig.clustering.VertexCover(OG, clusters=list(coverDict.values()))


def vertex_betweeenness_from_eb(G, eb):
    """
    Calculates the vertex betweenness scores in G. Returns a list
    in which the indices are the vertex indices and the values are
    their betweeennesses. The same as G.betweenness(), but faster because
    it uses the edge betweenness scores.
    (CONGA, page 4, equation 1)
    """
    components = G.components()
    membership = components.membership
    vbs = []
    for vertex in G.vs:
        numComponents = len(components[membership[vertex.index]])
        incidentEdges = G.incident(vertex)
        vb = 0.5 * (sum(G.es[e]["eb"] for e in incidentEdges) - (numComponents - 1))
        vbs.append(vb)
    return vbs


def get_neighborhood_vertex(G, v, h):
    """
    Given a vertex and a height/depth to
    traverse, find the neighborhood as defined in the CONGA
    paper.
    """
    return G.neighborhood(v, order=h)


def get_neighborhood_edge(G, e, h):
    """
    Given an edge and a height/depth to
    traverse, find the neighborhood as defined in the CONGA
    paper.
    """
    neigh = set(G.neighborhood(e[0], order=h - 1))
    neigh.update(G.neighborhood(e[1], order=h - 1))
    return list(neigh)


def order_tuple(toOrder):
    if toOrder[0] <= toOrder[1]:
        return toOrder
    return (toOrder[1], toOrder[0])


def create_clique(G, v, pb):
    """
    Given a vertex and its pair betweennesses, returns a k-clique
    representing all of its neighbors, with edge weights determined by the pair
    betweenness scores. Algorithm discussed on page 5 of the CONGA paper.
    """
    neighbors = G.neighbors(v)

    # map each neighbor to its index in the adjacency matrix
    mapping = {neigh: i for i, neigh in enumerate(neighbors)}
    n = len(neighbors)

    # Can use ints instead: (dtype=int). Only works if we use matrix_min
    # instead of mat_min.
    clique = np.zeros((n, n))
    for uw, score in pb.items():
        clique[mapping[uw[0]], mapping[uw[1]]] = score
        clique[mapping[uw[1]], mapping[uw[0]]] = score

    # Ignore any self loops if they're there. If not, this line
    # does nothing and can be removed.
    np.fill_diagonal(clique, 0)
    return clique


def reduce_matrix(M):
    """
    Given a matrix M, collapses the row and column of the minimum value. This is just
    an adjacency matrix way of implementing the greedy "collapse" discussed in CONGA.
    Returns the new matrix and the collapsed indices.
    """
    i, j = mat_min(M)
    # i, j = matrix_min(M)
    # add the ith row to the jth row and overwrite the ith row with those values
    M[i, :] = M[j, :] + M[i, :]

    # delete the jth row
    M = np.delete(M, (j), axis=0)

    # similarly with the columns
    M[:, i] = M[:, j] + M[:, i]
    M = np.delete(M, (j), axis=1)
    np.fill_diagonal(M, 0)  # not sure necessary.
    return i, j, M


def check_for_split(G, edge):
    """
    Given an edge in tuple form, check if it splits the
    graph into two disjoint clusters. If so, it returns
    True. Otherwise, False.
    """
    # Possibly keep a record of splits.
    try:
        return not G.edge_disjoint_paths(source=edge[0], target=edge[1])
        # TODO: specify exception (when edge = target)
    except Exception as e:
        return False


def mat_min(M):
    """
    Given a matrix, find an index of the minimum value (not including the
    diagonal).
    """
    # take a matrix we pass in, and fill the diagonal with the matrix max. This is
    # so that we don't grab any values from the diag.
    np.fill_diagonal(M, float("inf"))

    # figure out the indices of the cell with the lowest value.
    i, j = np.unravel_index(M.argmin(), M.shape)
    np.fill_diagonal(M, 0)
    return i, j


def matrix_min(mat):
    """
    Given a symmetric matrix, find an index of the minimum value
    in the upper triangle (not including the diagonal.)
    """
    # Currently, this function is unused, as its result is
    # the same as that of mat_min, and it is not always
    # faster. Left in for reference in case mat_min becomes
    # a bottleneck.

    # find the minimum from the upper triangular matrix
    # (not including the diagonal)
    upperTri = np.triu_indices(mat.shape[0], 1)
    minDex = mat[upperTri].argmin()

    # find the index in the big matrix. TODO: do so
    # with some algebra.
    triN = mat.shape[0] - 1
    row = 0

    while minDex >= triN:
        minDex -= triN
        triN -= 1
        row += 1
    col = mat.shape[0] - triN + minDex
    return row, col


def Congo_(graph, number_communities=0, height=2):
    """

    :param graph:
    :param number_communities:
    :param height:The lengh of the longest shortest paths that CONGO considers
    :return:
    """

    result = congo(graph, height)
    if number_communities == 0:
        cover = result._covers[result.optimal_count]
        number_communities = result.optimal_count
    else:
        cover = result._covers[number_communities]

    list_communities = []
    for i in range(0, number_communities):
        list_communities.append(cover._clusters[i])

    return list_communities



import seaborn as sns
import pandas as pd
import cdlib
import matplotlib.pyplot as plt
from typing import Callable

__all__ = [
    "plot_com_properties_relation",
    "plot_com_stat",
    "plot_sim_matrix",
    "plot_scoring",
]


def plot_sim_matrix(
    clusterings: list, scoring: Callable[[object, object], object]
) -> object:
    """
    Plot a similarity matrix between a list of clusterings, using the provided scoring function.

    :param clusterings: list of clusterings to compare
    :param scoring: the scoring function to use
    :return: a ClusterGrid instance

    Example:

    >>> from cdlib import algorithms, viz, evaluation
    >>> import networkx as nx
    >>> g = nx.karate_club_graph()
    >>> coms = algorithms.louvain(g)
    >>> coms2 = algorithms.walktrap(g)
    >>> clustermap = viz.plot_sim_matrix([coms,coms2],evaluation.adjusted_mutual_information)

    """
    forDF = []
    for c in clusterings:
        cID = c.get_description()
        for c2 in clusterings:
            c2ID = c2.get_description()
            forDF.append([cID, c2ID, scoring(c, c2).score])
    df = pd.DataFrame(columns=["com1", "com2", "score"], data=forDF)
    df = df.pivot("com1", "com2", "score")
    return sns.clustermap(df)


def plot_com_stat(
    com_clusters: list, com_fitness: Callable[[object, object, bool], object]
) -> object:
    """
    Plot the distribution of a property among all communities for a clustering, or a list of clusterings (violin-plots)

    :param com_clusters: list of clusterings to compare, or a single clustering
    :param com_fitness: the fitness/community property to use
    :return: the violin-plots

    Example:

    >>> from cdlib import algorithms, viz, evaluation
    >>> import networkx as nx
    >>> g = nx.karate_club_graph()
    >>> coms = algorithms.louvain(g)
    >>> coms2 = algorithms.walktrap(g)
    >>> violinplot = viz.plot_com_stat([coms,coms2],evaluation.size)

    """
    if isinstance(com_clusters, cdlib.classes.clustering.Clustering):
        com_clusters = [com_clusters]

    allVals = []
    allNames = []
    for c in com_clusters:
        prop = com_fitness(c.graph, c, summary=False)
        allVals += prop
        allNames += [c.get_description()] * len(prop)

    ax = sns.violinplot(allNames, allVals, cut=0, saturation=0.5, palette="Set3")
    for tick in ax.get_xticklabels():
        tick.set_rotation(90)

    plt.ylabel("%s" % com_fitness.__name__)
    plt.xlabel("Algorithm")
    plt.tight_layout()

    return ax


def plot_com_properties_relation(
    com_clusters: object,
    com_fitness_x: Callable[[object, object, bool], object],
    com_fitness_y: Callable[[object, object, bool], object],
    **kwargs: dict
) -> object:
    """
    Plot the relation between two properties/fitness function of a clustering

    :param com_clusters:  clustering(s) to analyze (cluster or cluster list)
    :param com_fitness_x: first fitness/community property
    :param com_fitness_y: first fitness/community property
    :param kwargs: parameters for the seaborn lmplot
    :return: a seaborn lmplot

    Example:

    >>> from cdlib import algorithms, viz, evaluation
    >>> import networkx as nx
    >>> g = nx.karate_club_graph()
    >>> coms = algorithms.louvain(g)
    >>> coms2 = algorithms.walktrap(g)
    >>> lmplot = viz.plot_com_properties_relation([coms,coms2],evaluation.size,evaluation.internal_edge_density)
    """
    if isinstance(com_clusters, cdlib.classes.clustering.Clustering):
        com_clusters = [com_clusters]

    for_df = []

    for c in com_clusters:
        x = com_fitness_x(c.graph, c, summary=False)
        y = com_fitness_y(c.graph, c, summary=False)
        for i, vx in enumerate(x):
            for_df.append([c.get_description(), vx, y[i]])

    df = pd.DataFrame(
        columns=["Method", com_fitness_x.__name__, com_fitness_y.__name__], data=for_df
    )
    ax = sns.lmplot(
        x=com_fitness_x.__name__,
        y=com_fitness_y.__name__,
        data=df,
        hue="Method",
        fit_reg=False,
        legend=False,
        x_bins=100,
        **kwargs
    )
    plt.legend(loc="best")
    plt.tight_layout()

    return ax


def plot_scoring(
    graphs: list,
    ref_partitions: object,
    graph_names: list,
    methods: list,
    scoring: Callable[
        [object, object], object
    ] = cdlib.evaluation.adjusted_mutual_information,
    nbRuns: int = 5,
) -> object:
    """
    Plot the scores obtained by a list of methods on a list of graphs.

    :param graphs: list of graphs on which to make computations
    :param ref_partitions: list of reference clusterings corresponding to graphs
    :param graph_names: list of the names of the graphs to display
    :param methods: list of functions that take a graph as input and return a Clustering as output
    :param scoring: the scoring function to use, default anmi
    :param nbRuns: number of runs to do for each method on each graph
    :return: a seaborn lineplot

    Example:

    >>> from cdlib import algorithms, viz, evaluation
    >>> import networkx as nx
    >>> g1 = nx.algorithms.community.LFR_benchmark_graph(1000, 3, 1.5, 0.5, min_community=20, average_degree=5)
    >>> g2 = nx.algorithms.community.LFR_benchmark_graph(1000, 3, 1.5, 0.7, min_community=20, average_degree=5)
    >>> names = ["g1", "g2"]
    >>> graphs = [g1, g2]
    >>> for g in graphs:
    >>>     references.append(NodeClustering(communities={frozenset(g.nodes[v]['community']) for v in g}, graph=g, method_name="reference"))
    >>> algos = [algorithms.crisp_partition.louvain, algorithms.crisp_partition.label_propagation]
    >>> viz.plot_scoring(graphs, references, names, algos, nbRuns=2)

    """
    forDF = []
    for i, g in enumerate(graphs):
        for m in methods:
            for r in range(nbRuns):
                partition = m(g)

                score = scoring(partition, ref_partitions[i]).score
                forDF.append([graph_names[i], score, partition.get_description()])
    df = pd.DataFrame(columns=["graph", "score", "method"], data=forDF)
    ax = sns.lineplot(x="graph", y="score", hue="method", data=df, legend="brief")
    ax.legend(loc="best")
    for tick in ax.get_xticklabels():
        tick.set_rotation(90)
    plt.tight_layout()

    return ax



# Author: True Price <jtprice@cs.unc.edu>

# A core-attachment based method to detect protein complexes in PPI networks
# Wu, Li, Kwoh, Ng (2009)
# http://www.biomedcentral.com/1471-2105/10/169

from collections import defaultdict
from itertools import combinations
import functools


# return average degree and density for a graph
def __graph_stats(graph):
    avg_deg = sum(len(n) for n in graph.values()) / float(len(graph))
    density = avg_deg / (len(graph) - 1)
    return avg_deg, density


# return core nodes, given a graph and its average degree
__get_core_nodes = lambda g, avg: set(v for v, n in g.items() if len(n) >= avg)

# return NA score
__NA_score = lambda a, b: float(len(a & b) ** 2) / (len(a) * len(b))


def __core_removal(graph, density_threshold):
    if len(graph) == 1:  # need at least two nodes in the graph...
        return [graph]

    avg_deg, density = __graph_stats(graph)

    if density >= density_threshold:
        return [graph]
    else:
        # find and remove core nodes; create connected subcomponents
        core_nodes = __get_core_nodes(graph, avg_deg)
        result = []
        subgraphs = []
        for v, n in graph.items():
            if v in core_nodes:
                continue
            n = n - core_nodes  # note that we're reassigning n
            for s in subgraphs:
                if not n.isdisjoint(s):
                    s |= n
                    break
            else:
                subgraphs.append(n | {v})
        # connected subcomponent joining
        i = 0
        while i < len(subgraphs) - 1:
            j = i + 1
            while j < len(subgraphs):
                if not subgraphs[i].isdisjoint(subgraphs[j]):
                    subgraphs[i] |= subgraphs[j]
                    subgraphs.pop(j)
                else:
                    j += 1
            i += 1
        # recursive core removal
        for s in subgraphs:
            tresults = __core_removal(
                dict((v, graph[v] & s) for v in s), density_threshold
            )
            for tc in tresults:
                nodes = set()
                for v, n in tc.items():
                    nodes.add(v)
                    n |= graph[v] & core_nodes
                for c in core_nodes:
                    tc[c] = graph[c] & (nodes | core_nodes)
            result += tresults
        return result


def co_ach(g, density_threshold=0.7, affinity_threshold=0.225, closeness_threshold=0.5):
    # read protein-protein pairs
    data = defaultdict(set)
    for a, b in g.edges():
        data[a].add(b)
        data[b].add(a)

    # step 1: find preliminary cores
    SC = []  # currently-detected preliminary cores
    count = 0
    for vertex, neighbors in data.items():
        # build neighborhood graph
        vertices = {vertex} | neighbors
        size1_neighbors = set()
        graph = {}
        for v in vertices:
            n = data[v] & vertices
            if len(n) > 1:  # ignore size-1 vertices
                graph[v] = n
            else:
                size1_neighbors.add(v)
        if len(graph) < 2:  # not enough connections in this graph
            continue
        graph[vertex] -= size1_neighbors

        # get core graph
        avg_deg, density = __graph_stats(graph)
        core_nodes = __get_core_nodes(graph, avg_deg)
        vertices = set(graph.keys())
        for v in vertices - core_nodes:
            del graph[v]
        for n in graph.values():
            n &= core_nodes
        if len(graph) < 2:  # not enough connections in this graph
            continue
        graph_nodes = set(graph)

        # inner loop
        for sg in __core_removal(graph, density_threshold):
            while True:
                _, density = __graph_stats(sg)
                # if density threshold met, stop; else, remove min degree node
                if density >= density_threshold:
                    break
                w = min(sg.items(), key=lambda k: len(k[1]))[0]
                del sg[w]
                for n in sg.values():
                    n.discard(w)

            sg_nodes = set(sg)
            while graph_nodes - sg_nodes:
                w = max(graph_nodes - sg_nodes, key=lambda v: len(graph[v] & sg_nodes))
                new_sg = sg.copy()
                for v, n in new_sg.items():
                    if w in graph[v]:
                        n.add(w)
                new_sg[w] = graph[w] & sg_nodes
                _, density = __graph_stats(new_sg)
                if density < density_threshold:
                    break
                sg = new_sg
                sg_nodes.add(w)

            # redundancy filtering
            max_sim = -1
            for i in range(len(SC)):
                sim = __NA_score(set(SC[i]), sg_nodes)
                if sim > max_sim:
                    max_sim = sim
                    index = i
            if max_sim < affinity_threshold:
                SC.append(sg)
            else:
                _, density_i = __graph_stats(SC[index])
                if density * len(sg) > density_i * len(SC[index]):
                    SC[index] = sg

    # step 2: adding peripheral proteins
    clusters = set()
    for core in SC:
        nodes = frozenset(core)
        neighbors = (
            functools.reduce(lambda x, y: x | y, (data[v] for v in nodes)) - nodes
        )
        neighbors -= set(
            v
            for v in neighbors
            if float(len(data[v] & nodes)) / len(nodes) <= closeness_threshold
        )
        clusters.add(nodes | neighbors)

    return [list(c) for c in clusters]



import collections
import numpy as np
import networkx as nx
from sklearn.cluster import DBSCAN


class WalkSCAN(object):
    def __init__(self, nb_steps=2, eps=0.1, min_samples=3):
        self.nb_steps = nb_steps
        self.eps = eps
        self.min_samples = min_samples
        self.dbscan_ = DBSCAN(eps=self.eps, min_samples=self.min_samples)

    def load(self, graph, init_vector):
        self.graph = graph.copy()
        self.init_vector = init_vector.copy()

    def embed_nodes(self):
        p = {0: self.init_vector.copy()}
        for t in range(self.nb_steps):
            p[t + 1] = collections.defaultdict(int)
            for v in p[t]:
                for (_, w, e_data) in self.graph.edges(v, data=True):
                    if "weight" in e_data:
                        self.weighted_ = True
                        p[t + 1][w] += (
                            float(e_data["weight"])
                            / float(self.graph.degree(v, weight="weight"))
                            * p[t][v]
                        )
                    else:
                        self.weighted_ = False
                        p[t + 1][w] += 1.0 / float(self.graph.degree(v)) * p[t][v]
        self.embedded_value_ = dict()
        self.embedded_nodes_ = list()
        for v in p[self.nb_steps]:
            self.embedded_nodes_.append(v)
            self.embedded_value_[v] = np.array(
                [p[t + 1][v] for t in range(self.nb_steps)]
            )
        self.nb_embedded_nodes_ = len(self.embedded_nodes_)

    def find_cores(self):
        if self.nb_embedded_nodes_ > 0:
            P = np.zeros((self.nb_embedded_nodes_, self.nb_steps))
            for (i, node) in enumerate(self.embedded_nodes_):
                P[i, :] = self.embedded_value_[node]
            self.dbscan_.fit(P)
            self.cores_ = collections.defaultdict(set)
            self.outliers_ = set()
            for (i, node) in enumerate(self.embedded_nodes_):
                label = self.dbscan_.labels_[i]
                if label >= 0:
                    self.cores_[label].add(node)
                else:
                    self.outliers_.add(node)
        else:
            self.cores_ = {}
            self.outliers_ = set()

    def compute_core_average_value(self):
        self.core_average_value_ = dict()
        for (core_id, core) in self.cores_.items():
            self.core_average_value_[core_id] = np.zeros(self.nb_steps)
            for node in core:
                for t in range(self.nb_steps):
                    self.core_average_value_[core_id][t] += self.embedded_value_[node][
                        t
                    ] / float(len(core))

    def sort_cores(self):
        self.sorted_core_ids_ = list(self.cores_.keys())
        self.sorted_core_ids_.sort(
            key=lambda i: list(self.core_average_value_[i]), reverse=True
        )
        self.sorted_cores_ = [self.cores_[i] for i in self.sorted_core_ids_]

    def aggregate_outliers(self):
        self.communities_ = list()
        for core in self.sorted_cores_:
            community = core.copy()
            for node in core:
                community |= set(nx.neighbors(self.graph, node)) & self.outliers_
            self.communities_.append(community)

    def detect_communities(self, graph, init_vector):
        self.load(graph, init_vector)
        self.embed_nodes()
        self.find_cores()
        self.compute_core_average_value()
        self.sort_cores()
        self.aggregate_outliers()



from cdlib.classes.node_clustering import NodeClustering
from cdlib import evaluation


class AttrNodeClustering(NodeClustering):

    """Attribute Node Communities representation.

    :param communities: list of communities
    :param graph: a networkx/igraph object
    :param method_name: community discovery algorithm name
    :param coms_labels: dictionary specifying for each community the frequency of the attribute values
    :param method_parameters: configuration for the community discovery algorithm used
    :param overlap: boolean, whether the partition is overlapping or not
    """

    def __init__(
        self,
        communities: list,
        graph: object,
        method_name: str = "",
        coms_labels: dict = None,
        method_parameters: dict = None,
        overlap: bool = False,
    ):
        super().__init__(communities, graph, method_name, method_parameters, overlap)
        self.coms_labels = coms_labels

    def purity(self) -> evaluation.FitnessResult:

        """Purity is the product of the frequencies of the most frequent labels carried by the nodes within the communities
        :return: FitnessResult object
        """
        res = None
        if self.coms_labels is not None:
            res = evaluation.purity(self.coms_labels)
        return res



from cdlib import TemporalClustering, NamedClustering
from cdlib.algorithms.internal_dcd.eTILES import eTILES

__all__ = ["tiles"]


def tiles(dg: object, obs: int = 1) -> TemporalClustering:
    """
    TILES is designed to incrementally identify and update communities in stream graphs.
    This implementation assume an explicit edge removal when pairwise interactions cease to exist.


    **Supported Graph Types**

    ========== ======== ========
    Undirected Directed Weighted
    ========== ======== ========
    Yes        No       No
    ========== ======== ========

    :param dg: dynetx graph object
    :param obs: community observation interval (default=1)
    :return: TemporalClustering object

    :Example:

    >>> from cdlib import algorithms
    >>> import dynetx as dn
    >>> dg = dn.DynGraph()
    >>> for x in range(10):
    >>>     g = nx.erdos_renyi_graph(200, 0.05)
    >>>     dg.add_interactions_from(list(g.edges()), t=x)
    >>> coms = algorithms.tiles(dg, 2)

    :References:

    Rossetti, Giulio; Pappalardo, Luca; Pedreschi, Dino, and Giannotti, Fosca. `Tiles: an online algorithm for community discovery in dynamic social networks.<https://link.springer.com/article/10.1007/s10994-016-5582-8>`_ Machine Learning (2016), 106(8), 1213-1241.
    """
    alg = eTILES(dg=dg, obs=obs)
    tc = TemporalClustering()
    t = obs
    for c in alg.execute():
        communities = {}
        for k, v in c.items():
            communities[f"{t}_{k}"] = v
        sg = dg.time_slice(t - obs, t)

        nc = NamedClustering(communities, sg, "TILES", {"obs": obs}, overlap=True)

        if t <= max(dg.temporal_snapshots_ids()):
            tc.add_clustering(nc, time=t)
            t += obs
        else:
            break

    # add community matches (can contain communities not present in the observations)
    mtc = alg.get_matches()
    tc.add_matching(mtc)

    # cleaning & updating community matching
    dg = tc.lifecycle_polytree(None, False)
    community_ids = list(dg.nodes())

    tids = tc.get_observation_ids()
    ndss = []
    for tid in tids:
        c = tc.get_clustering_at(tid)
        comunity_ids_m = c.named_communities.keys()
        for ci in comunity_ids_m:
            ndss.append(ci)

    for tid in community_ids:
        if tid not in ndss:
            dg.remove_node(tid)

    mtc = list(dg.edges())
    tc.add_matching(mtc)

    return tc
